#!/bin/sh

###############################################################################
#
#   File:     astroroot_setup.sh
#
#   !!! Do not change the layout of the following line !!!
    ASRO_SETUP_VERSION=1.0
#
#   Component: astroroot_init_env
#
#   Author(s): Mathias.Beck@obs.unige.ch (MB)  (isdc_init_env.csh)
#	       Reiner.Rohlfs   version 1.0 astroroot_init_env.csh
#
#   Purpose:   script to set up a file to initialise AstroROOT relevant
#              environment variables
#
#   Revision History:
#
#      Version 1.0:  modified from isdc_init_env.csh version 0.4.3
#
###############################################################################

ASRO_SETUP_HELP=${ISDC_ENV}/help/astroroot_init_env.txt


########################################
#
# astroroot_setup_log
#
########################################

astroroot_setup_log() {

   log_type=$1
   log_mess=$2
   
   if [ ${log_type} = "e" ] ; then
      echo "Error  : "${log_mess}
   elif [ ${log_type} = "w" ] ; then
      echo "Warning: "${log_mess}
   else
      echo "Log    : "${log_mess}
   fi
}

########################################
#
# astroroot_setup_help
#
########################################

astroroot_setup_help() {

   if [ -r ${ASRO_SETUP_HELP} ] ; then
      /bin/cat ${ASRO_SETUP_HELP}
   else
      exit 4
   fi
}


########################################
#
# astroroot_setup_main
#
########################################


####################
#
# Check the number of arguments. 
# Exactly one argument is supported
#
####################

if [ $# -ne 1 ] ; then
   exit 5
fi

if [ $1 = "--version" ] ; then
   astroroot_setup_log "l" "astroroot_init_env ${ASRO_SETUP_VERSION}"
   exit 0
elif [ $1 = "--help" ] ; then
   astroroot_setup_help
   exit 0
fi


####################
#
# Only sh and csh flavours are supported
#
####################

if [ $1 = "sh" ] ; then
   flavour="sh"
elif [ $1 = "csh" ] ; then
   flavour="csh"
else
   exit 6
fi



####################
#
# Check the setting of ROOTSYS
#
####################

if [ x$ROOTSYS = "x" ] ; then
   return_code=42
fi


####################
#
# set the variables
#
####################


####################
#
# PATH
#
####################

if [ x$ROOTSYS != "x" ] ; then
  if [ -d ${ROOTSYS}/bin ]; then
    ROOT_BIN=${ROOTSYS}/bin

    # prune existing references to ISDC_BIN from path, then put
    # ISDC_BIN at the front. extra colons are so that match
    # will occur anywhere in path
    PATH=`echo ":${PATH}:" | sed "s%:${ROOT_BIN}:%:%g" | sed "s%::*$%%"`
    PATH="${ROOT_BIN}${PATH}"
  fi
fi
if [ -d ${ISDC_ENV}/bin ]; then
  ASRO_BIN=${ISDC_ENV}/bin

  # prune existing references to ASRO_BIN from path, then put
  # ASRO_BIN at the front. extra colons are so that match
  # will occur anywhere in path
  PATH=`echo ":${PATH}:" | sed "s%:${ASRO_BIN}:%:%g" | sed "s%::*$%%"`
  PATH="${ASRO_BIN}${PATH}"
fi


####################
#
# MAN_PATH
#
####################

if [ x$ROOTSYS != "x" ] ; then
  if [ -d ${ROOTSYS}/man ]; then
    ROOT_MAN=${ROOTSYS}/man

    # prune existing references to ROOT_MAN from man path, then put
    # ROOT_MAN at the front. extra colons are so that match
    # will occur anywhere in path
    MANPATH=`echo ":${MANPATH}:" | sed "s%:${ROOT_MAN}:%:%g" | sed "s%::*$%%"`
    MANPATH="${ROOT_MAN}${MANPATH}"
  fi
fi


####################
#
# LD_LIBRARY_PATH
#
####################

if [ x$ROOTSYS != "x" ] ; then
  if [ -d ${ROOTSYS}/lib ]; then
    ROOT_LIB=${ROOTSYS}/lib

    # prune existing references to ROOT_LIB from path, then put
    # ROOT_LIB at the front. extra colons are so that match
    # will occur anywhere in path
    LD_LIBRARY_PATH=`echo ":${LD_LIBRARY_PATH}:" | sed "s%:${ROOT_LIB}:%:%g" | sed "s%::*$%%"`
    LD_LIBRARY_PATH="${ROOT_LIB}${LD_LIBRARY_PATH}"
  fi
fi

if [ -d ${ISDC_ENV}/lib ]; then
  ASRO_LIB=${ISDC_ENV}/lib

  # prune existing references to ASRO_LIB from path, then put
  # ASRO_LIB at the front. extra colons are so that match
  # will occur anywhere in path
  LD_LIBRARY_PATH=`echo ":${LD_LIBRARY_PATH}:" | sed "s%:${ASRO_LIB}:%:%g" | sed "s%::*$%%"`
  LD_LIBRARY_PATH="${ASRO_LIB}${LD_LIBRARY_PATH}"
fi


####################
#
# PFILES
#
####################

if [ -d ${ISDC_ENV}/pfiles ]; then

   # make sure local parameter file directory exists
   if [ ! -w ${HOME}/pfiles/. ]; then
      mkdir ${HOME}/pfiles
      if [ $? -ne 0 ]; then
      	 exit 3
      fi
   fi

   # use PFILES (if set) to set initial values for locpfiles and syspfiles
   if [ "x${PFILES}" != x ]; then
      syspfiles=`echo ${PFILES} | sed "s%.*;%%"`
      locpfiles=`echo ${PFILES} | sed "s%;*${syspfiles}$%%"`
   fi

   # move HOME/pfiles to front of locpfiles and weed out redundancies
   if [ "x${locpfiles}" = x ]; then
      locpfiles="${HOME}/pfiles"
   else
      locpfiles=`echo ":${locpfiles}:" | \
      	 sed "s%:${HOME}/pfiles:%:%g" | sed "s%::*$%%"`
      locpfiles="${HOME}/pfiles${locpfiles}"
   fi

   # move ISDC_ENV/pfiles to front of syspfiles and weed out redundancies
   if [ "x${syspfiles}" = x ]; then
      syspfiles="${ISDC_ENV}/pfiles"
   else
      syspfiles=`echo ":${syspfiles}:" | \
      	 sed "s%:${ISDC_ENV}/pfiles:%:%g" | sed "s%::*$%%"`
      syspfiles="${ISDC_ENV}/pfiles${syspfiles}"
   fi

   PFILES="${locpfiles};${syspfiles}"
fi


var_list="PATH MANPATH LD_LIBRARY_PATH PFILES"


####################
#
# save values in output initialization shell scripts
#
####################

tmp_astroroot_config=${HOME}/.astroroot_config.$$

if [ ${flavour} = "sh" ]; then
   #
   # Bourne Shell version, write only for sh flavour initialisation
   #
  rm -f ${tmp_astroroot_config}.sh
  touch ${tmp_astroroot_config}.sh
  for var in ${var_list} ; do
    eval "if [ x\$$var != x ]; then
            echo \"\$var=\\\"\$$var\\\"; export \$var\" >> ${tmp_astroroot_config}.sh;
          fi"
  done
  echo "${tmp_astroroot_config}.sh"
elif [ ${flavour} = "csh" ]; then
   #
   # C-Shell version, write only for csh flavour initialisation
   #
  rm -f ${tmp_astroroot_config}.csh
  touch ${tmp_astroroot_config}.csh
  for var in ${var_list} ; do
    eval "if [ x\$$var != x ]; then
            echo setenv \$var \\\"\$$var\\\" >> ${tmp_astroroot_config}.csh;
          fi"
  done
  echo "${tmp_astroroot_config}.csh"
fi
exit ${return_code}
