/* $Id: keyword_cache.c,v 1.5 2002/04/26 13:26:15 oneel Exp $ */

/* $Log: keyword_cache.c,v $
/* Revision 1.5  2002/04/26 13:26:15  oneel
/* Fix SPR1247, use bsearch rather than my buggy binary search
/*
/* Revision 1.4  2002/04/11 13:02:49  oneel
/* cfitiso 2.201.7 release.
/*
/* Fixes SCREW 512.  If one is just modifying the value of a keyword
/* don't nuke the cache, just change the cache value.
/*
/* Revision 1.3  2002/03/22 13:23:20  oneel
/* Small fix to keyword_cache.c so that in the special case where the hdu
/* is positioned beyond the current number of keywords we don't overwrite
/* memory anymore.  In this case we assume that the number of keywords in
/* the hdu are the same as the current position in the hdu.
/*
/* Revision 1.2  2002/03/19 13:40:15  oneel
/* Change strncpy to strcpy.  Strncpy zero fills the output buffer upto
/* the length n.  Honestly I don't know how long the output buffer is
/* (and the lenght is not passed in as it should be) so 0 filling to some
/* arbritary length is not right.  Therefore all the strncpy instances
/* were replaced with strcpy.
/* */
/* a file of keyword caching routines to speed up cfitiso, in fact it is 
   supposed to speed up DAL.  */

/* files in the base cfitsio modified
   fitsio.h - changed FITSfile structure
   fitiso2.h - added prototypes
   Makefile.in - added keyword_cache.c
   cfileio.c - routine ffclos needs to free the cache.
               routines ffomem, ffopen, ffinit, ffimem all needed to 
	       init the cache pointer array.
   getkey.c - ffgcrd needs to call ffget_cached_key_value.
   buffers.c - the routines which make buffers dirty need to clear the 
   cache as well.  These are ffpbyt and ffpbytoff.  

   Changes to clear the cache less, ie, just when a keyword changes

   putkey.c - routine ffprec.
   edithdu.c - ffcpdt for the outfptr.
   modkey.c - ffikey, ffdrec
   fitscore.c - ffmkey, ffwend

   and as a result we removed the calls from buffers.c to clear the cache

   added keyword_cache.c (this file)
*/

/* This is meant to be a black box from the rest of cfitsio, if for no
   other reason that we might want to change the implimentation a
   number of times.  The exported routines are:

   int ffget_cached_key_value( fitsfile *fptr,  char *name, char *card, 
        int *status);
   int ffclear_key_cache_selected( fitsfile *fptr);
   int ffclear_key_cache( fitsfile *fptr);
   int ffprint_cache_stats(void);

   ffget_cached_key_value is called from ffgcrd and mimics that routine
   to get a cached fits header card.  It also (to mimic the semantics of
   ffgcrd) moves the next fits card position for reading to what ever
   offset the returned card would have left the reading position at.

   ffclear_key_cache is called to clear ALL of the caches.  This should
   be called in ffclos etc.

   ffclear_key_cache_selected removes all of the caches from the
   current hdu on.  Because we cache the position in the fits file
   where this keyword was stored we must clear AT LEAST from the
   current HDU on.  It's possible that this will be teh same as
   ffclear_key_cache in the end and one can not assume that this is
   doesn't clear all of the caches.  Normally this is called in ffpbyt 
   and ffpbytoff though it seems for less overhead it should be called
   from ffmrec, ffmcrd, ffmkyx, ffmkyu etc.  This is not done.
   

   ffprint_cache_stats just prints some handy cache stats

*/


#include "fitsio2.h"
#include <stdlib.h>
#include <strings.h>
#include <stdio.h>
#include <assert.h>

#define REAL_CARD_LEN 88
#define REAL_KEYWORD_LEN 12

static int cache_build_in_progress = 0;

static int cache_builds = 0;
static int cache_clears = 0;
static int cache_searches = 0;
static int cache_hits = 0;
static int cache_misses = 0;
static int cache_updates = 0;

void ffprint_cache_stats(void)
{
    printf("Cache stats:\nBuilds: %d Clears: %d\nSearch %d Hit %d Miss %d Updates %d\n",
	   cache_builds,cache_clears,cache_searches,cache_hits,cache_misses,
	   cache_updates);
}

	   

typedef struct    /* struct for one key cache entry */
{
    int cardnum;
    char keyword[REAL_KEYWORD_LEN];
    char card[REAL_CARD_LEN]; 
  /* Why not a constant, like say FLEN_LEN?
     Because we need this to be long word 
     alligned (for some archs) because we
     are going to access these as a big block of
     memory for speed */
  
} cache_card;

typedef struct    /* struct to hold one HDUs worth of cache */
{
  int numcards;
  void *cache;

} hdu_cache;


/* the entry fptr->Fptr->cached_header_data can be either:
   - Null
   - An array of MAXHDU pointers.  Each of these pointers 
   can be either NULL or point to one hdu_cache structure.  All of this has
   to be freed.
*/

static int clear_key_cache(fitsfile *fptr, int start)
{
    /* Clear a cache starting at hdu start */
    hdu_cache *tmp_cache;
    int ii;
    
    if (start > MAXHDU) return 0;

    assert(cache_build_in_progress == 0);
    
#ifdef DEBUG
    printf ("Clear cache called with %d\n",start);
#endif
    fflush(stdout);
  
    for (ii=start;ii<MAXHDU;ii++) {
      tmp_cache = (hdu_cache *) fptr->Fptr->cached_header_data[ii];

      if (tmp_cache) {
	if (tmp_cache->cache) free(tmp_cache->cache);
	free(tmp_cache);
	cache_clears++;

      }
      fptr->Fptr->cached_header_data[ii] = NULL;
    
    }
    return 0;
}

int ffclear_key_cache( fitsfile *fptr)
{
    /* clears the keyword cache for file fptr */
    return clear_key_cache(fptr,1);
}

int ffclear_key_cache_selected( fitsfile *fptr)
{
    
    /* Clear the cache from the current hdu */
    return clear_key_cache(fptr,fptr->Fptr->curhdu);
}

static int cache_compare(const void *p1, const void *p2)
     /* sorting comparision function */

{
  cache_card *c1 = (cache_card *) p1;
  cache_card *c2 = (cache_card *) p2;
  return strncasecmp(c1->keyword, c2->keyword, REAL_KEYWORD_LEN);
}
  
static int search_compare(const void *p1, const void *p2)
     /* searching comparision function */
     
{
  char *c1 = (char *) p1;
  cache_card *c2 = (cache_card *) p2;
  int retval;
  
  retval = strncasecmp(c1, c2->keyword, REAL_KEYWORD_LEN);
  return retval;
}

static int build_cache (fitsfile *fptr, int *status)
{

    /* Build a cache */
    int nexist, position;
    hdu_cache *tmp_cache;
    int ii, ntodo, jj, kk;
#ifdef TESTDEBUG
    char card[100];
#endif
    int cardlen;
    char cardname[FLEN_KEYWORD];
    
    cache_card *current_cache_card;
        
    cache_build_in_progress = 1;
    cache_builds++;
    /* printf("Building cache for %s\n",fptr->Fptr->filename); */
            
    ffghps(fptr,&nexist,&position,status);
    /* for some odd reason, position can be beyond nexist.  This doesn't 
       seem correct but when creating a new header this seems to happen.  
       so, let's prevent this from being a problem...  This was SPR 1118.
    */
    if (position > nexist) nexist = position;

    fptr->Fptr->cached_header_data[fptr->Fptr->curhdu] =
	(hdu_cache *)calloc(1,sizeof(hdu_cache));
    
    tmp_cache = ((hdu_cache *)fptr->Fptr->cached_header_data
	[fptr->Fptr->curhdu]);

    if (!tmp_cache) {
	/* Memory allocation error */
	*status = MEMORY_ALLOCATION;
	cache_build_in_progress = 0;
	return *status;
    }
    
    tmp_cache->numcards = nexist;
    tmp_cache->cache = calloc(nexist,sizeof(cache_card));
    /* You did catch what happend there, right?  We are going to 
       access our 'array' of cache cards as a big void pointer
       block of memory.  Othewise, if we did a **cache_card we would 
       have to call calloc/malloc 1000s of time for a big header.  
       malloc calls take time. */

    if (!tmp_cache->cache) {
	*status = MEMORY_ALLOCATION;
	cache_build_in_progress = 0;
	return *status;
    }

    /* The next bit of code is copied straight out of cfitsio/getkey.c/ffgcrd
       so that we read the keys just like cfitsio does.  */


    ntodo = nexist - position + 1;  /* first, read from next keyword to end */
    for (jj=0; jj < 2; jj++)
	{
	    for (kk = 0; kk < ntodo; kk++)
		{
		    ffghps(fptr,&nexist,&ii,status);
		    current_cache_card = 
			(cache_card *)((char *)tmp_cache->cache + 
				       ((ii-1)*sizeof(cache_card)));
		    ffgnky(fptr, ((cache_card *)current_cache_card)->card,
			   status);     /* get next keyword */
		    /* get the keyword name */
		    ffgknm(((cache_card *)current_cache_card)->card,
			   cardname, &cardlen, status); 
		    strcpy(((cache_card *)current_cache_card)->keyword,
			    cardname);
		    ((cache_card *)current_cache_card)->
			keyword[REAL_KEYWORD_LEN-1] = '\0';
		    ((cache_card *)current_cache_card)->cardnum = ii;
                }
	    /* reset pointer to beginning of header */
	    ffmaky(fptr, 1, status);  
	    ntodo = position - 1;      /* number of keyword to read */ 
	}
    


    /* Now sort it */

    qsort(tmp_cache->cache, nexist, sizeof(cache_card), cache_compare);
    for (ii=1;ii<=nexist;ii++) {
      current_cache_card = (cache_card *)((char *)tmp_cache->cache + 
					  ((ii-1)*sizeof(cache_card)));
    }
    *status = ffmaky(fptr, position, status);
    cache_build_in_progress = 0;
    return *status;
}

static int fetch_cached_value (fitsfile *fptr,  
			       char *name, char *card, 
			       cache_card ** retval,
			       int *status)
{
  /* fetch the cached value */

  hdu_cache * tmp_cache;
  cache_card *current_cache_card;
  int low, high, mid;
  int cmp;
  
  tmp_cache = ((hdu_cache *)fptr->Fptr->cached_header_data
    [fptr->Fptr->curhdu]); 

  if (!tmp_cache) {
    /* well, some error here, there is no cache */
      *status = MEMORY_ALLOCATION;
      return *status;
  }
  
  if (!(tmp_cache->cache)) {
      /* still no cache */
      *status = MEMORY_ALLOCATION;
      return *status;
  }
  
  /* low = 0;
     high = tmp_cache->numcards - 1;
     while ((low <= high) && !*retval) {
     mid = (low + high)/2;
     current_cache_card = 
     (cache_card *)((char *)tmp_cache->cache + 
     ((mid-1)*sizeof(cache_card)));
     cmp = strncasecmp(name, current_cache_card->keyword,REAL_KEYWORD_LEN);
     if (cmp < 0)
     high = mid - 1;
     else if (cmp > 0)
     low = mid + 1;
     else
     *retval = current_cache_card;
     }
  */
  *retval = bsearch(name, tmp_cache->cache,tmp_cache->numcards,
     sizeof(cache_card), search_compare);
   
  cache_searches++;
  /* did we find it? */
  if (*retval) {
    /* move to the right place with ffmaky */
      *status = ffmaky(fptr,(*retval)->cardnum+1,status); 
      strcpy(card,(*retval)->card);
      *status = 0;
      cache_hits++;
    return *status;
  }
  /* printf("Trying to find and did not: %s file %s\n",name,fptr->Fptr->filename); */
  cache_misses++;
  *status = KEY_NO_EXIST;
  return *status; /* we didn't find it */

}


int ffget_cached_key_value( fitsfile *fptr,  
			    char *name, char *card, 
			    int *status)
    /* This routine fetches the cached value, after possibly building
       a new cache.  Normally it is ONLY called from ffgcrd.  The
       semantics SHOULD be identical, just this should be faster.
       NOTE: Wildcarded keyword searches are not searched here, so
       they don't speed up.  It's possible that they could be
       depending on the caching method but don't depend on it.
    */
{

  cache_card * retval = NULL;
    
  /* Do we have a cache for the current hdu? */
  if (!fptr->Fptr->cached_header_data[fptr->Fptr->curhdu])
    {
      /* Make the cache */
#ifndef CACHE_DEBUG
      build_cache (fptr, status);
#endif
      
    }
  
  /* Now fetch it */
  
  if (!*status) {
    fetch_cached_value (fptr, name, card, &retval, status);
  }
  
  return *status;
  
}

int ffmodify_key_cache(fitsfile *fptr, char * tcard, int *status)
     /* Modify the keyword cache to update the tcard given.  
	What we'll do here is figure out what keyword we are 
	changing, fetch the value out of the cache, and update
	the returned value.  We ASSUME that the position in the 
	header of the keyword did NOT change.  Ie, this is an in
	place update.
     */
	
{
  
  int cardlen;
  char cardname[FLEN_KEYWORD];
  char card[REAL_CARD_LEN];
  cache_card * retval = NULL;

  
  if (*status > 0)           /* inherit input status value if > 0 */
    return(*status);

  /* Which keyword */
  ffgknm(tcard, cardname, &cardlen, status);

  if (*status > 0) return(*status);
  
  /* Do we have a cache for the current hdu? */
  if (!fptr->Fptr->cached_header_data[fptr->Fptr->curhdu])
    {
      /* This is odd, one would expect a cache here given that we just 
	 recently fetched some value.  Still, I don't think building 
	 the cache here is a good idea */
      return (*status);
    }
  
  /* Now fetch it */
  
  if (!*status) {
    fetch_cached_value (fptr, cardname, card, &retval, status);
  }

  if ((*status > 0) || (0 != strcmp(retval->keyword,cardname))) {
    /* well, that means that either some cfitiso routine is using
       a low level routine, or, something very very bad has happend.
       If it's not a low level cfitiso routine getting called I'll 
       guess that either the cache is screwed or there is some 
       internal logic error.  In this case just clear the cache
       and return, we aren't going to succeed */
    *status = 0;
    return ffclear_key_cache_selected(fptr);
  }
  
  strcpy(retval->card,tcard);
  cache_updates++;
  return *status;
}
