/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                        PARAMETER - GUI                              */
/*                                                                     */
/*                                                                     */
/*  File:      param_gui_main.cpp                                      */
/*  Version:   1.0                                                     */
/*  Component: param_gui                                               */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   A program to edit FTOOLS parameter files with a GUI     */
/*             The param_gui_lib library is used to perform the task.  */
/*             Here is just the main program and some support          */
/*             functions.                                              */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 14/08/02 RR  First version                            */
/*                                                                     */
/***********************************************************************/
#include <unistd.h>
#include <time.h>

#include <TROOT.h>
#include <TApplication.h>
#include <TGMsgBox.h>

#include "pil.h"

#include "param_gui.h"



class ProgSupport : public Support
{
public:
   ProgSupport(const char * paraFileName, const char * progName, 
               const char * guiProgName, T_Widgets * widgets)
      : m_progName(progName), m_guiProgName(guiProgName)
      {m_paraFileName = paraFileName; m_widgets = widgets;}

   virtual const char * GetParamFileName() {return m_paraFileName;}
   virtual void   Save(const char * filename = NULL);
   virtual void   Terminate() 
                     {gApplication->Terminate(0);}
   virtual void   Run();

private:

   void  SaveAs(const char * filename);

   T_Widgets  * m_widgets;
   const char * m_paraFileName;
   std::string  m_progName;
   std::string  m_guiProgName;

};

/*****************************************************************************/
/*  Function :     ProgSupport::Run                                          */
/*                                                                           */
/*  Description:   This function is called when the user press the Run -     */
/*                 button. It starts the assigned program with the current   */
/*                 parameters                                                */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*****************************************************************************/
void ProgSupport::Run() 
{
   std::string comment = m_progName + " ";

   for (I_Widgets i_widget = m_widgets->begin();
        i_widget != m_widgets->end(); i_widget++)
      comment += (*i_widget)->GetRunComment();
      
   comment += "&";

   printf("%s\n", comment.c_str() );

   system(comment.c_str() );
}

/*****************************************************************************/
/*  Function :     ProgSupport::Save                                         */
/*                                                                           */
/*  Description:   This function is called when the user press the Save or   */
/*                 SaveAs - button to store the current parameters in the    */
/*                 parameter file.                                           */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  filename    input                    if != NULL the parameters are saved */
/*                                       in a new file, else in the          */
/*                                       original parameter file is updated  */
/*                                                                           */
/*****************************************************************************/
void ProgSupport::Save(const char * filename)
{
   if (filename != NULL && filename[0] != 0)
      {
      SaveAs(filename);
      return;
      }

   for (I_Widgets i_widget = m_widgets->begin();
        i_widget != m_widgets->end(); i_widget++)
      (*i_widget)->UpdateParamFile();
    
   PILFlushParameters();     
}

/*****************************************************************************/
/*  Function :     ProgSupport::SaveAs                                       */
/*                                                                           */
/*  Description:   This function is called when the user press the SaveAs    */
/*                 button. It writes the parameters and the GUI attributes   */
/*                 into a new parameter file.                                */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  filename    input   != NULL          filename of new parameter file      */
/*                                                                           */
/*****************************************************************************/
void ProgSupport::SaveAs(const char * filename)
{
   // test if the file already exist
   if (!access(filename, F_OK))
      {
      char msg[600];
      Int_t ret;
      sprintf(msg, "File %s exist\noverwrite?", filename);
      new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                  "File Exist", msg,
                   kMBIconExclamation, kMBYes | kMBNo, &ret);
      if (ret == kMBNo)
         return;
      unlink(filename);
      }

   // open the file
   FILE * fl = fopen(filename, "w");
   if (fl == NULL)
      {
      char msg[600];
      sprintf(msg, "Cannot open file\n %s", filename);
      new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                  "File Open Error", msg,
                   kMBIconExclamation, kMBOk);
      return;
      }

   // write header
   fprintf(fl, "#\n# This file is created by the program\n# %s.\n#\n", 
                m_guiProgName.c_str());
   fprintf(fl, "# It is a modified FTOOLS parameter file of\n");
   fprintf(fl, "# %s\n#\n", m_paraFileName);
   time_t now = time(NULL);
   fprintf(fl, "# Creation time: %s#\n#\n", ctime(&now));

   // write common attributes
   Widget::WriteCommonAttr(fl);

   // write parameters
   for (I_Widgets i_widget = m_widgets->begin();
        i_widget != m_widgets->end(); i_widget++)
      (*i_widget)->WriteParamFile(fl);


   // write widget attributes
   for (I_Widgets i_widget = m_widgets->begin();
        i_widget != m_widgets->end(); i_widget++)
      (*i_widget)->WriteWidgetAttr(fl);

   fclose(fl);
   
}

/*****************************************************************************/
/*  Function :     main                                                      */
/*                                                                           */
/*  Description:   main function of this program                             */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  argc        input                    number of program parameters        */
/*  argv        input                    array of program parameters         */
/*                                                                           */
/*****************************************************************************/
main(int argc, char ** argv)
{

   int tmpArgc = 1;	// to avoid changing the current directroy
   TApplication theApp("param_gui", &tmpArgc, argv);


   // there has to be at least 1 program parameter
   if (argc < 2)
      {
      new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                  "Error Message", 
                  "Program parameter is missing!\n"
                  "usage: param_gui parameterfile.par [program name]",
	                kMBIconExclamation, kMBOk);
      return 1;
      }

   // initialze PIL
   // set mode in PIL to be able to overwrite parameters
   PILSpecialMode |= PIL_PSET_MODE;
   PILSetModuleName(argv[1]);
   int status = PILInit(argc, argv);

   if (status != ISDC_OK)
      {
      char msg[240];
      sprintf(msg, "Cannot find and open the parameter file.\n"
                  "%s\nPIL returns error %d", argv[1], status);
      new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                  "Error Message", msg,
                  kMBIconExclamation, kMBOk);
      return 1;
      }


   // get from PIL the parameter file name
   char * paraFileName;
   PILGetParFilename(&paraFileName);

   // this is our list of all widgets
   T_Widgets widgets;

   // prepare the progname
   char progName[512];
   if (argc == 2)
      {
      char * pos = strrchr(paraFileName, '/');
      if (pos)
         strcpy(progName, pos + 1);
      else
         strcpy(progName, paraFileName);
      pos = strstr(progName, ".par");
      if (pos)
         *pos = 0;
      }
   else
      strcpy(progName, argv[2]);

   // create the support function class for this run
   ProgSupport support(paraFileName, progName, argv[0], &widgets);


   // open the parameter file and create the widgets
   FILE * file = fopen(paraFileName, "r");
   if (file == NULL)
      return 1;

   char line[PIL_LINESIZE];

   while (fgets(line, PIL_LINESIZE, file) != NULL)
      {
      // filter out EOL characters
      char * p;
      if (NULL != (p = strchr(line, '\n'))) *p = 0;
      if (NULL != (p = strchr(line, '\r'))) *p = 0;

      ParseLine(line, widgets);
      }

   fclose(file);

   // everything is prepared open the first window
   PG_MainWindow * mainWindow = new PG_MainWindow(&widgets, &support); 

   // run the program and accept user input
   theApp.Run(kTRUE);

   PILClose(ISDC_OK);
   
   delete mainWindow;
}

