/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*  File:      main_window.cpp                                         */
/*  Version:   1.0                                                     */
/*  Component: param_gui_lib                                           */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   The main window of the param - gui. Should be called    */
/*             only once.                                              */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 14/08/02 RR  First version                            */
/*                                                                     */
/***********************************************************************/
#include <TGMsgBox.h>
#include <TGFileDialog.h>
#include <TGXYLayout.h>

#include "param_gui.h"
#include "tools.h"
#include "widget.h"

T_Widgets * PG_MainWindow::m_widgets;
Support   * PG_MainWindow::m_support;



/*****************************************************************************/
/*  Function :     PG_MainWindow::PG_MainWindow                              */
/*                                                                           */
/*  Description:   Constructor of the main window.                           */
/*                 Depending on the size and positions of the widgets        */
/*                 this window is resized.                                   */
/*                 The widgets which shall be placed on this main window     */
/*                 are attached to it.                                       */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  widget      input                    list of all widgets as read from    */
/*                                       the parameter file with all         */
/*                                       # GUI - attributes                  */
/*  support     input                    interface class to able to call     */
/*                                       some functions of the calling       */
/*                                       program                             */
/*                                                                           */
/*****************************************************************************/
PG_MainWindow::PG_MainWindow(T_Widgets * widgets, Support * support)
  : TGMainFrame(gClient->GetRoot(), 200, 50)
{	

   m_widgets = widgets;
   m_support = support;

   StdButton stdButton(0);
   WinSize   winSize(stdButton);


   for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
     (*i_w)->Init();

   // set the window size depending of the defined size of
   // the widgets
   for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
      winSize.SetSize(*i_w, "main");

   // set the positon of the other widgets and than set the
   // window size depending of the size of these widgets
   winSize.SetMode(1);
   for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
      winSize.SetSize(*i_w, "main");

   // create the hidden button - widget
   Widget * hidden = new W_Window("_HIDDEN_");
   hidden->SetAttribute("NAME:hidden");
   char attr[20];
   sprintf(attr, "X:%05.1f", winSize.GetWidth() - stdButton.GetWidth());
   hidden->SetAttribute(attr);
   sprintf(attr, "Y:%05.1f", stdButton.GetHeight() - 3);
   hidden->SetAttribute(attr);
   m_widgets->push_back(hidden);

   Resize((UInt_t)(winSize.GetWidth() + 1.0), (UInt_t)(winSize.GetHeight() + 1.0));

   SetLayoutManager(new TGXYLayout(this));

   // draw all widgets, first the frames
   for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
     (*i_w)->CreateFrame(this, "main");
   for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
     (*i_w)->Create(this, "main");

   // draw the standard buttons
   stdButton.Create(this, &winSize);

   MapSubwindows();
   Layout();
   SetWindowName(m_support->GetParamFileName());
   MapWindow();

   for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
     (*i_w)->BeforeOpen(string("main"));

}
/*****************************************************************************/
/*  Function :     PG_MainWindow:: ~PG_MainWindow                            */
/*                                                                           */
/*  Description:   Destructor, deletes all widgets (in memory and on the     */
/*                 screen                                                    */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
PG_MainWindow:: ~PG_MainWindow()
{
  for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
    delete *i_w;

  TGFrameElement *ptr;

  // delete all frames and layout hints
  if (fList)
    {
    TIter next(fList);
    while ((ptr = (TGFrameElement *) next())) 
      {
      delete ptr->fLayout;
      delete ptr->fFrame;
      }
   }

}
/*****************************************************************************/
/*  Function :     PG_MainWindow::ProcessMessage                             */
/*                                                                           */
/*  Description:   Standard ROOT function to be called as a result of a      */
/*                 user event (button of the mouse was pressed)              */
/*                                                                           */
/*  Return code:   kTRUE                                                     */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  See ROOT user manual                                                     */
/*                                                                           */
/*****************************************************************************/
Bool_t PG_MainWindow::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{

   if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == BT_QUIT)
      {
/*      if (Widget::GetModifiedFlag())
         {
         Int_t rc;
         new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                      "Save Data?", 
                      "Modified Parameters are not saved!\nSave before closing GUI?",
    	                 kMBIconQuestion, kMBYes | kMBNo | kMBCancel, &rc);
         if (rc == kMBCancel)
            return kTRUE;
         if (rc == kMBYes)
            {
            m_support->Save();
            Widget::ClearModifiedFlag();
            }
         }
*/
      CloseWindow();
      }

   else if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == BT_SAVE)
      {
      m_support->Save();
      Widget::ClearModifiedFlag();
      }

   else if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == BT_SAVE_AS)
      {
      TGFileInfo fi;
      const char * fileTypes[]  = {"parameter file", "*.par", 0, 0};
      fi.fFileTypes = fileTypes; 
      new TGFileDialog(gClient->GetRoot(), gClient->GetRoot(),
                       kFDSave, &fi);
      if (fi.fFilename == NULL)
         // the user pressed the cancle button
         return kTRUE;

      // make sure that the filename ends with .par
      char fileName[512];
      strcpy(fileName, fi.fFilename);
      if (strcmp(fileName + strlen(fileName) - 4, ".par"))
         strcat(fileName, ".par");
      
      m_support->Save(fileName);
      Widget::ClearModifiedFlag();
      }

    else if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == BT_RUN)
      m_support->Run();

    else if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == BT_HELP)
       OpenHelp(this, m_support->GetParamFileName());

   else
      for (I_Widgets i_w = BeginW(); i_w != EndW(); i_w++)
        (*i_w)->ProcessMessage(msg, param1, param2);

   return kTRUE;
}
/*****************************************************************************/
/*  Function :     PG_MainWindow::CloseWindow                                */
/*                                                                           */
/*  Description:   Called to close this window. This function calls a        */
/*                 callback function of the calling program which can        */
/*                 decide what to do.                                        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void PG_MainWindow::CloseWindow()
{
   m_support->Terminate();   
}




