/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*  File:      param_api.cpp                                           */
/*  Version:   1.2                                                     */
/*  Component: param_gui_lib                                           */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   Implementation of a application class. This class can   */
/*             be used by applications to open a GUI with all          */
/*             parameters and to get the values of all parameters      */
/*             set by the user.                                        */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.1: 29/05/03 RR  First version                            */
/*          1.2: 14/11/03 RR  ParamGui::ParamGui has new function      */
/*                            paraemter: app                           */
/*                                                                     */
/***********************************************************************/
#include <unistd.h>
#include <time.h>

#include <TROOT.h>
#include <TApplication.h>
#include <TGMsgBox.h>

#include "pil.h"
#include "param_api.h"


/*****************************************************************************/
/*  Function :     ParamGui::Api_Support::Save                               */
/*                                                                           */
/*  Description:   This function is called when the user press the Save or   */
/*                 SaveAs - button to store the current parameters in the    */
/*                 parameter file.                                           */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  filename    input                    if != NULL the parameters are saved */
/*                                       in a new file, else the             */
/*                                       original parameter file is updated  */
/*                                                                           */
/*****************************************************************************/
void ParamGui::Api_Support::Save(const char * filename)
{
   if (filename != NULL && filename[0] != 0)
      {
      SaveAs(filename);
      return;
      }

   for (I_Widgets i_widget = m_widgets.begin();
        i_widget != m_widgets.end(); i_widget++)
      (*i_widget)->UpdateParamFile();
    
   PILFlushParameters();     

}
/*****************************************************************************/
/*  Function :     ParamGui::Api_Support::SaveAs                             */
/*                                                                           */
/*  Description:   This function is called when the user press the SaveAs    */
/*                 button. It writes the parameters and the GUI attributes   */
/*                 into a new parameter file.                                */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  filename    input   != NULL          filename of new parameter file      */
/*                                                                           */
/*****************************************************************************/
void ParamGui::Api_Support::SaveAs(const char * filename)
{
   // test if the file already exist
   if (!access(filename, F_OK))
      {
      char msg[600];
      Int_t ret;
      sprintf(msg, "File %s exist\noverwrite?", filename);
      new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                  "File Exist", msg,
                   kMBIconExclamation, kMBYes | kMBNo, &ret);
      if (ret == kMBNo)
         return;
      unlink(filename);
      }

   // open the file
   FILE * fl = fopen(filename, "w");
   if (fl == NULL)
      {
      char msg[600];
      sprintf(msg, "Cannot open file\n %s", filename);
      new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                  "File Open Error", msg,
                   kMBIconExclamation, kMBOk);
      return;
      }

   // write header
   fprintf(fl, "#\n# This file is created by the program\n# %s.\n#\n", 
                m_progName);
   fprintf(fl, "# It is a modified FTOOLS parameter file of\n");
   fprintf(fl, "# %s\n#\n", m_paramFileName);
   time_t now = time(NULL);
   fprintf(fl, "# Creation time: %s#\n#\n", ctime(&now));

   // write common attributes
   Widget::WriteCommonAttr(fl);

   // write parameters
   for (I_Widgets i_widget = m_widgets.begin();
        i_widget != m_widgets.end(); i_widget++)
      (*i_widget)->WriteParamFile(fl);


   // write widget attributes
   for (I_Widgets i_widget = m_widgets.begin();
        i_widget != m_widgets.end(); i_widget++)
      (*i_widget)->WriteWidgetAttr(fl);

   fclose(fl);
   
}
/*****************************************************************************/
/*  Function :     ParamGui::Api_Support::Run                                */
/*                                                                           */
/*  Description:   This function is called when the user press the Run -     */
/*                 button. It saves the actual definition of all parameters  */
/*                 in m_result and terminates the GUI.                       */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*****************************************************************************/
void ParamGui::Api_Support::Run()
{
   // copy the user input into the result map
   for (I_Widgets i_widget = m_widgets.begin();
        i_widget != m_widgets.end(); i_widget++)
      m_result[(*i_widget)->GetName()] = (*i_widget)->GetValue();

   m_quit = false; 
   gApplication->Terminate(0);
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/*****************************************************************************/
/*  Function :     ParamGui::ParamGui                                        */
/*                                                                           */
/*  Description:   Constructor of this class. It opens the GUI and returns   */
/*                 after the user closed the GUI either with the QUIT button */
/*                 or the Run button.                                        */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  argc        input                    number of program parameters        */
/*  argv        input                    array of program parameters         */
/*  app         input   my be NULL       this application is used if not NULL*/
/*                                       if NULL a new one is created.       */
/*                                                                           */
/*****************************************************************************/
ParamGui::ParamGui(int argc, char ** argv, TApplication * app)
{
   m_error = ISDC_OK;
   Bool_t  internalApp = kFALSE;

   if (app == NULL)
      {
      int tmpArgc = 1;	// to avoid changing the current directory
      app = new TApplication("param_gui", &tmpArgc, argv);
      internalApp = kTRUE;
      }

   // initialize PIL
   // set mode in PIL to be able to overwrite parameters
   PILSpecialMode |= PIL_PSET_MODE;
   int status = PILInit(argc, argv);
   if (status != ISDC_OK)
      {
      m_error = status;
      return;
      }

   // get from PIL the parameter file name and set the program name
   PILGetParFilename(&m_support.m_paramFileName);
   m_support.m_progName = argv[0];

   // open the parameter file and create the widgets
   FILE * file = fopen(m_support.m_paramFileName, "r");
   if (file == NULL)
      {
      m_error = -1;
      PILClose(ISDC_OK);
      return;
      }

   char line[PIL_LINESIZE];

   while (fgets(line, PIL_LINESIZE, file) != NULL)
      {
      // filter out EOL characters
      char * p;
      if (NULL != (p = strchr(line, '\n'))) *p = 0;
      if (NULL != (p = strchr(line, '\r'))) *p = 0;

      ParseLine(line, m_support.m_widgets);
      }

   fclose(file);


   // everything is prepared open the first window
   PG_MainWindow * mainWindow = new PG_MainWindow(&m_support.m_widgets, &m_support); 

   // run the program and accept user input
   app->Run(kTRUE);

   PILClose(ISDC_OK);

   delete mainWindow;

   if (internalApp)
      delete app;

}
/*****************************************************************************/
/*  Function :     ParamGui::GetBool                                         */
/*                                                                           */
/*  Description:   Returns the value of one BOOL parameter. This function    */
/*                 should not be called if the user pressed the QUIT button: */
/*                 the method Quit() return true!                            */
/*                                                                           */
/*  Return code:   the value of the requested parameter                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  paraName    input                    the name of the parameter for       */
/*                                       which the user defined value will   */
/*                                       be returned.                        */
/*                                                                           */
/*****************************************************************************/
bool ParamGui::GetBool(const char * paraName)
{
   std::map<std::string, std::string>::iterator i_result =
      m_support.m_result.find(paraName);
   if (i_result != m_support.m_result.end())
      return i_result->second == "yes";

   m_error = -2;
   return false;
}
/*****************************************************************************/
/*  Function :     ParamGui::GetInt                                          */
/*                                                                           */
/*  Description:   Returns the value of one Int parameter. This function     */
/*                 should not be called if the user pressed the QUIT button: */
/*                 the method Quit() return true!                            */
/*                                                                           */
/*  Return code:   the value of the requested parameter                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  paraName    input                    the name of the parameter for       */
/*                                       which the user defined value will   */
/*                                       be returned.                        */
/*                                                                           */
/*****************************************************************************/
int ParamGui::GetInt(const char * paraName)
{
   std::map<std::string, std::string>::iterator i_result =
      m_support.m_result.find(paraName);
   if (i_result != m_support.m_result.end())
      return atoi(i_result->second.c_str());

   m_error = -2;
   return 0;
}
/*****************************************************************************/
/*  Function :     ParamGui::GetReal                                         */
/*                                                                           */
/*  Description:   Returns the value of one Real parameter. This function    */
/*                 should not be called if the user pressed the QUIT button: */
/*                 the method Quit() return true!                            */
/*                                                                           */
/*  Return code:   the value of the requested parameter                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  paraName    input                    the name of the parameter for       */
/*                                       which the user defined value will   */
/*                                       be returned.                        */
/*                                                                           */
/*****************************************************************************/
double ParamGui::GetReal(const char * paraName)
{
   std::map<std::string, std::string>::iterator i_result =
      m_support.m_result.find(paraName);
   if (i_result != m_support.m_result.end())
      {
      double res;
      sscanf(i_result->second.c_str(), "%lf", &res);
      return res;
      }

   m_error = -2;
   return 0.0;
}
/*****************************************************************************/
/*  Function :     ParamGui::GetString                                       */
/*                                                                           */
/*  Description:   Returns the value of one string parameter. This function  */
/*                 should not be called if the user pressed the QUIT button: */
/*                 the method Quit() return true!                            */
/*                                                                           */
/*  Return code:   the value of the requested parameter                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  paraName    input                    the name of the parameter for       */
/*                                       which the user defined value will   */
/*                                       be returned.                        */
/*                                                                           */
/*****************************************************************************/
const char * ParamGui::GetString(const char * paraName)
{
   std::map<std::string, std::string>::iterator i_result =
      m_support.m_result.find(paraName);
   if (i_result != m_support.m_result.end())
      return i_result->second.c_str();

   m_error = -2;
   static const char str[1] = "";
   return str;
}

