/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*                                                                     */
/*  File:      param_gui.h                                             */
/*  Version:   1.0                                                     */
/*  Developer: Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 14/08/02 RR  First version                            */
/*                                                                     */
/***********************************************************************/
#ifndef PARAM_GUI_INCLUDED
#define PARAM_GUI_INCLUDED

#include <list>
#include <set>
#include <string>

#include <TGFrame.h>


class Widget;


typedef std::list<Widget*>          T_Widgets;
typedef T_Widgets::const_iterator   I_Widgets;


class Support
{
public:
   virtual const char * GetParamFileName() = 0;
   virtual void   Save(const char * filename = NULL) = 0;
   virtual void   Terminate() = 0;
   virtual void   Run() = 0;
};


class PG_MainWindow : public TGMainFrame
{
public:
   PG_MainWindow(T_Widgets * widgets, Support * support );
   ~PG_MainWindow();

   Bool_t ProcessMessage(Long_t msg, Long_t param1, Long_t param2);
   virtual void CloseWindow();

   static I_Widgets  BeginW() {return m_widgets->begin();}
   static I_Widgets  EndW()   {return m_widgets->end();}

   static const char * GetParamFileName() {return m_support->GetParamFileName();}

private:
  
  static  T_Widgets   * m_widgets;
  static  Support     * m_support;

};


class Widget
{
public:
   Widget(const char * name);

   virtual ~Widget() {}

   virtual void SetAttribute(const char * attrStr);
   static  void SetCommonAttribute(const char * attrStr);

   virtual void MakeSize() {}

   virtual void Init()  {}   // called before Create at start of program

   // Create called for each new window
   virtual void Create(TGCompositeFrame * win, const char * winName) {}
   virtual void CreateFrame(TGCompositeFrame * win, const char * winName) {}

   // called just before opening a new window
   virtual void BeforeOpen(const std::string & backGround) {} 

   virtual void ProcessMessage(Long_t msg, Long_t param1, Long_t param2) {}
   virtual void CloseWidget(std::string & backGround) 
                     {if (backGround == m_z) m_open = false;}
   virtual void ValueChanged(std::string & widgetName, std::string & value) {}

   static void   ClearModifiedFlag()    {g_modified = kFALSE;}
   static Bool_t GetModifiedFlag()      {return g_modified;}

   bool   SizeDefined(){return (m_width > 0.0) && (m_height > 0.0);}
   bool   PosDefined() {return (m_xPos >= 0.0) && (m_yPos >= 0.0);}
   double GetXPos()    {return m_xPos;}
   double GetYPos()    {return m_yPos;}
   double GetWidth()   {return m_width + m_nameW;}
   double GetHeight()  {return m_height;}
   const std::string & GetZ()      {return m_z;}
   const std::string & GetName()   {return m_name;}
   const std::string & GetValue()  {return m_value;}

   virtual void   SetXPos(double xPos)     {m_xPos = xPos;}
   virtual void   SetYPos(double yPos)     {m_yPos = yPos;}
   virtual void   SetWidth(double width)   {m_width = width;}
   virtual void   SetHeight(double height) {m_height = height;}

   virtual void SetParamType(const char * paramType) {m_paramType = paramType;}
   virtual void SetMode     (const char * mode);
   virtual void SetValue    (const char * value)     {m_value     = value;}
   virtual void SetMin      (const char * min)       {m_min       = min;}
   virtual void SetMax      (const char * max)       {m_max       = max;}
   virtual void SetPrompt   (const char * prompt)    {m_prompt    = prompt;}

   virtual std::string GetRunComment(); 
   virtual void        UpdateParamFile()  {}
   virtual void        WriteParamFile(FILE * fl);
   virtual void        WriteWidgetAttr(FILE * fl);
   static  void        WriteCommonAttr(FILE * fl);

protected:
   void CreateName(TGCompositeFrame * win, double yOff);

   static Bool_t g_modified;

   static Int_t  g_id;  // unique id counter
          Int_t  m_id;  // id of this widget

   bool   m_open;  // true if the widget is visible on its background

   std::string   m_name;
   std::string   m_paramType;
   std::string   m_mode;
   std::string   m_value;
   std::string   m_min;
   std::string   m_max;
   std::string   m_prompt;

   double   m_xPos, m_yPos;
   double   m_width, m_height;

   static double  g_textW;
   static double  g_fileW;
   static double  g_nameW;
   double         m_nameW;

   std::string    m_guiName;   // name only for the GUI
   std::string    m_tabNames;  // list of all tabs for a folder widget

   std::string    m_z; // name of the background frame

   int            m_type;    
   std::string    m_fitsExt;   // default extension in FITS file

   std::string             m_disableW;  // disable depends on this widget
   std::set<std::string>   m_disableV;  // disable this widget if m_disableW has
                              // one of these values
   bool         m_disabled;   // true if this widget is disabled
};


void ParseLine(const char * line, T_Widgets & widgets);


#endif // PARAM_GUI_INCLUDED












