/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*  File:      w_combobox.cpp                                          */
/*  Version:   1.0                                                     */
/*  Component: param_gui_lib                                           */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   The combobox widget. This is used if the min field of   */
/*             the parameter is a list of accepted values or strings   */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 16/08/02 RR  First version                            */
/*                                                                     */
/***********************************************************************/

#include <TGComboBox.h>
#include <TGXYLayout.h>

#include "tools.h"
#include "widget.h"
#include "param_gui.h"

#include "pil.h"


/*****************************************************************************/
/*  Function :     W_ComboBox::W_ComboBox                                    */
/*                                                                           */
/*  Description:   A combobox is used if the min field of the parameter is   */
/*                 a list of accepted values. There has to be at least one   */
/*                 '|' character in the min field                            */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     the parameter name                  */
/*  type       input    s,f,i and r      paraemter type                      */
/*                                                                           */
/*****************************************************************************/
W_ComboBox::W_ComboBox(const char * name, char type)
  : Widget(name)
{
   m_height = 2.0;
   m_width  = 16;
   m_type   = 0;

   switch (type)
      {
      case 's':  m_type = T_STRING; break;
      case 'f':  m_type = T_FILE;   break;
      case 'i':  m_type = T_INT;    break;
      case 'r':  m_type = T_REAL;   break;
      }
}
/*****************************************************************************/
/*  Function :     W_ComboBox::SetMin                                        */
/*                                                                           */
/*  Description:   Set the min field of the parameter.                       */
/*                 Here the width of the widget is determined. It is the     */
/*                 length of the larges string plus 6                        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_ComboBox::SetMin (const char * min)
{
   Widget::SetMin(min);

   // set the width depending on the longest string
   int index0 = 0;
   int index1;
   int length = 0;
   while ((index1 = m_min.find('|', index0)) != -1)
      {
      if (index1 - index0 > length)
         length = index1 - index0;
      index0 = index1 + 1;
      }

   if ( m_min.size() - index0 > length)
      length =  m_min.size() - index0;

   m_width = length + 6;
}
/*****************************************************************************/
/*  Function :     W_ComboBox::Init                                          */
/*                                                                           */
/*  Description:   Initialize all other widgets (enable, disable them)       */
/*                 depending on the value of this widget. This is called     */
/*                 just after the widget classes with all attributes were    */
/*                 created, but before the screen widgets were created.      */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_ComboBox::Init()
{
   if (m_value.empty())
      return;

   for (I_Widgets i_w = PG_MainWindow::BeginW();
         i_w != PG_MainWindow::EndW(); i_w++)
      (*i_w)->ValueChanged(m_name, m_value);
}
/*****************************************************************************/
/*  Function :     W_ComboBox::Create                                        */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_ComboBox::Create(TGCompositeFrame * win, const char * winName)
{
   if (m_z != winName)
      return;

   CreateName(win, 0.2);

   m_comboBox = new TGComboBox(win, g_id);
   m_id = g_id;
   g_id++;

   // fill the combobox
   SubStr sub(m_min);
   string str;
   int count = 0;
   while(sub.Next(str))
      {
      m_comboBox->AddEntry(str.c_str(), count);
      if (str == m_value)
         m_comboBox->Select(count);
      count++;
      }

   win->AddFrame(m_comboBox, new TGXYLayoutHints(
         m_xPos + m_nameW, m_yPos + 0.1, m_width - 1, 1.8));

   m_win = win; 
   m_open = true;
}
/*****************************************************************************/
/*  Function :     W_ComboBox::ProcessMessage                                */
/*                                                                           */
/*  Description:   This function is called from the window to perform the    */
/*                 necessary action when the user hids this widget           */
/*                 The m_value is updated.                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  msg        input                     the ROOT message for widgets        */
/*  param1     input                     the id of the widget that was hit   */
/*                                       by the user. This function does     */
/*                                       nothing if param1 is not the id of  */
/*                                       this widget.                        */
/*  param2     input                     furter information, often not used  */
/*                                                                           */
/*****************************************************************************/
void W_ComboBox::ProcessMessage(Long_t msg, Long_t param1, 
                                Long_t param2)
{
   if (param1 != m_id || GET_SUBMSG(msg) != kCM_COMBOBOX)
      return;


   // find the selected string
   SubStr sub(m_min);
   string newVal;
   while(sub.Next(newVal))
      {
      if (param2 == 0)
         break;
      param2--;
      }

   if (newVal == m_value)
      return;

   m_value = newVal;
   g_modified = kTRUE;

   for (I_Widgets i_w = PG_MainWindow::BeginW();
         i_w != PG_MainWindow::EndW(); i_w++)
      (*i_w)->ValueChanged(m_name, m_value);

}
/*****************************************************************************/
/*  Function :     W_ComboBox::ValueChanged                                  */
/*                                                                           */
/*  Description:   This function is called from other widgets if they change */
/*                 their value. This widget can be enabled or disabled       */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  widgetName  input                    parameter name of the widget which  */
/*                                       changed its value                   */
/*  value       input                    new value of widgetName             */
/*                                                                           */
/*****************************************************************************/
void W_ComboBox::ValueChanged(string & widgetName, string & value)
{
   if (widgetName != m_disableW)   return;
   m_disabled = m_disableV.find(value) != m_disableV.end();

   if (m_open)
      {
      if (m_disabled)
         m_win->HideFrame(m_comboBox);
      else
         m_win->ShowFrame(m_comboBox);    
      }

}
/*****************************************************************************/
/*  Function :     W_ComboBox::BeforeOpen                                    */
/*                                                                           */
/*  Description:   This function is called after the widget is created on    */
/*                 its window, but before the window appears on the screen   */
/*                 The combobox is set to "hidden" if it is disabled.        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  backGround  input                    this is the window which informs    */
/*                                       this widget.                        */
/*                                                                           */
/*****************************************************************************/
void W_ComboBox::BeforeOpen(const string & backGround)
{
   if (backGround == m_z)
      if (m_disabled)
         m_win->HideFrame(m_comboBox);	
}
/*****************************************************************************/
/*  Function :     W_ComboBox::UpdateParamFile                               */
/*                                                                           */
/*  Description:   Writes its value into the parameter file using PIL        */
/*                 functions.                                                */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_ComboBox::UpdateParamFile()
{
   switch (m_type)
      {
      case T_STRING:
         PILPutString(m_name.c_str(), m_value.c_str());
         break;
      case T_FILE:
         PILPutFname(m_name.c_str(), m_value.c_str());
         break;
      case T_INT:
         if (!m_value.empty())
            PILPutInt(m_name.c_str(), atoi(m_value.c_str()) );
         break;
      case T_REAL:
         if (!m_value.empty())
            {
            double val;
            sscanf(m_value.c_str(), "%lf", &val);
            PILPutReal(m_name.c_str(), val );
            }
         break;
      default: ;
      }
}

