/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*  File:      w_folder.cpp                                            */
/*  Version:   1.0                                                     */
/*  Component: param_gui_lib                                           */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   A folder with tabs is implemented here.                 */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 17/08/02 RR  First version                            */
/*                                                                     */
/***********************************************************************/

#include <TGFrame.h>
#include <TGXYLayout.h>

#include "param_gui.h"
#include "widget.h"
#include "tools.h"


/*****************************************************************************/
/*  Function :     W_Folder::W_Folder                                        */
/*                                                                           */
/*  Description:   A folder with tabs. A parameter independent widget        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     the widget name, always _FOLDER_*   */
/*                                                                           */
/*****************************************************************************/
W_Folder::W_Folder(const char * name)
  : Widget(name)
{
  m_height = -1;
  m_width  = -1;
  m_nameW  = 0;
}
/*****************************************************************************/
/*  Function :     W_Folder::MakeSize                                        */
/*                                                                           */
/*  Description:   This function is called while the size of the window is   */
/*                 calculated. If asks every widget which should be in one   */
/*                 if the tabs for their size.                               */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_Folder::MakeSize()
{
  WinSize mainWinSize;

  I_Widgets begin = PG_MainWindow::BeginW();
  I_Widgets end   = PG_MainWindow::EndW();
  string str;
  SubStr sub(m_tabNames);
  while (sub.Next(str))
    {  
    WinSize winSize;
    
    // set the window size depending of the defined size of
    // the widgets
    for (I_Widgets i_w = begin; i_w != end; i_w++)
      winSize.SetSize(*i_w, str.c_str());

    // set the positon of the other widgets and than set the
    // window size depending of the size of these widgets
    winSize.SetMode(1);
    for (I_Widgets i_w = begin; i_w != end; i_w++)
      winSize.SetSize(*i_w, str.c_str());
    
    mainWinSize.Or(winSize);
    }

  m_width = mainWinSize.GetWidth() + 1;
  m_height = mainWinSize.GetHeight() + 2.5;

}
/*****************************************************************************/
/*  Function :     W_Folder::Create                                          */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_Folder::Create(TGCompositeFrame * win, const char * winName)
{
  if (m_z != winName)
    return;

  FolderWindow * folder = new FolderWindow(win, m_width -1, m_height - 0.5,
                                           m_tabNames);

  win->AddFrame(folder, new TGXYLayoutHints(
         m_xPos + 0.5, m_yPos + 0.25, m_width - 1, m_height - 0.5,
         TGXYLayoutHints::kLRubberX | TGXYLayoutHints::kLRubberY | 
         TGXYLayoutHints::kLRubberW | TGXYLayoutHints::kLRubberH));

  m_open = true;
}
/*****************************************************************************/
/*  Function :     W_Folder::CloseWidget                                     */
/*                                                                           */
/*  Description:   This widget will be closed. All widget in this folder     */
/*                 are informed that they will be closed.                    */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  backGround  input                    the name of the window that is      */
/*                                       closed. The name of the window      */
/*                                       is the z - coordinate of that       */
/*                                       window.                             */
/*                                                                           */
/*****************************************************************************/
void W_Folder::CloseWidget(string & backGround)
{
  Widget::CloseWidget(backGround);

  if (backGround != m_z)
    return;

  string str;
  SubStr sub(m_tabNames);
  while (sub.Next(str))
    {
    for (I_Widgets i_w = PG_MainWindow::BeginW(); 
         i_w != PG_MainWindow::EndW(); i_w++)
      (*i_w)->CloseWidget(str);
    }   

}
/*****************************************************************************/
/*  Function :     osm_gti                                                   */
/*                                                                           */
/*  Description:   This function is called after the widget is created on    */
/*                 its window, but before the window appears on the screen   */
/*                 All widgets in this folder are informed about this        */
/*                 status.                                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  backGround  input                    this is the window which informs    */
/*                                       this widget.                        */
/*                                                                           */
/*****************************************************************************/
void W_Folder::BeforeOpen(const string & backGround)
{
 
  if (backGround != m_z)
    return;

  string str;
  SubStr sub(m_tabNames);
  while (sub.Next(str))
    {
    for (I_Widgets i_w = PG_MainWindow::BeginW(); 
         i_w != PG_MainWindow::EndW(); i_w++)
      (*i_w)->BeforeOpen(str);
    }   

}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     FolderWindow::FolderWindow                                */
/*                                                                           */
/*  Description:   This is the folder widget on the screen. It creaets all   */
/*                 its tabs                                                  */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     parent frame. This frame will be    */
/*                                       located on this frame               */
/*  width      input                     widht of this folder                */
/*  height     input                     height of this folder               */
/*  tabNames   input                     all names ( = z - coordinate) of    */
/*                                       all tabs of this folder frame. They */
/*                                       are seperated by '|'                */
/*                                                                           */
/*****************************************************************************/
FolderWindow::FolderWindow(TGCompositeFrame * win, Double_t width,
                           Double_t height, string & tabNames)
  : TGTab(win, 10, 10)
{
  m_hints = new TGXYLayoutHints(0, 0, width, height - 2,
                                   TGXYLayoutHints::kLRubberW | TGXYLayoutHints::kLRubberH);

  FolderTab * tab;
  string str;
  SubStr sub(tabNames);
  while (sub.Next(str))
    {
    tab = new FolderTab(AddTab(str.c_str()), m_hints, str);
    m_tabs.insert(m_tabs.end(), tab); 
    }
 
  SetTab(0);
}
/*****************************************************************************/
/*  Function :     FolderWindow:: ~FolderWindow                              */
/*                                                                           */
/*  Description:   deletes all tabs and the layout hint.                     */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
FolderWindow:: ~FolderWindow()
{
  delete m_hints;

  list<FolderTab*>::iterator i_f = m_tabs.begin();
  while (i_f != m_tabs.end())
    {
    delete *i_f;
    i_f++;
    }
};

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     FolderTab::FolderTab                                      */
/*                                                                           */
/*  Description:   One tab of the folder above.                              */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  topWin      input                    the parent frame of this tab        */
/*  hints       input                    the layout hints for topWin         */
/*  name        input                    the name (= z - coordinate) of this */
/*                                       tab                                 */
/*                                                                           */
/*****************************************************************************/
FolderTab::FolderTab(TGCompositeFrame * topWin, 
                     TGXYLayoutHints * hints,
                     string & name)
  : TGCompositeFrame(topWin, 40, 100)
{
  topWin->SetLayoutManager(new TGXYLayout(topWin));
  SetLayoutManager(new TGXYLayout(this));

  I_Widgets begin = PG_MainWindow::BeginW();
  I_Widgets end   = PG_MainWindow::EndW();
  // draw all widgets into this tab, first the frames
  for (I_Widgets i_w = begin; i_w != end; i_w++)
    (*i_w)->CreateFrame(this, name.c_str());
  for (I_Widgets i_w = begin; i_w != end; i_w++)
    (*i_w)->Create(this, name.c_str());

  topWin->AddFrame(this, hints);
}
/*****************************************************************************/
/*  Function :     FolderTab::~FolderTab                                     */
/*                                                                           */
/*  Description:   Deletes all child widgets and their layout hints          */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
FolderTab::~FolderTab()
{
  TGFrameElement *ptr;

  // delete all frames and layout hints
  if (fList)
    {
    TIter next(fList);
    while ((ptr = (TGFrameElement *) next())) 
      {
      delete ptr->fLayout;
      delete ptr->fFrame;
      }
   }
}
/*****************************************************************************/
/*  Function :     FolderTab::ProcessMessage                                 */
/*                                                                           */
/*  Description:   Called from the ROOT inner loop when the user press the   */
/*                 mouse button. This message is passed to all widgets.      */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  See ROOT user manual                                                     */
/*                                                                           */
/*****************************************************************************/
Bool_t FolderTab::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{
  I_Widgets begin = PG_MainWindow::BeginW();
  I_Widgets end   = PG_MainWindow::EndW();
  int count = 0;
  for (I_Widgets i_w = begin; i_w != end; i_w++)
    (*i_w)->ProcessMessage(msg, param1, param2);
  return kTRUE; 
}


