/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*  File:      w_text.cpp                                              */
/*  Version:   1.3                                                     */
/*  Component: param_gui_lib                                           */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   A text widget to enter either just a text, a filename   */
/*             or selecting a data structure from a FITS file, a DOL   */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 19/08/02 RR  First version                            */
/*          1.3: 06/02/04 RR  remove dal dependency and use            */
/*                            TFIOElement                              */
/*                                                                     */
/***********************************************************************/
#include <TGTextEntry.h>
#include <TGTextBuffer.h>
#include <TGButton.h>
#include <TGListBox.h>
#include <TGFileDialog.h>
#include <TGMsgBox.h>
#include <TGXYLayout.h>

#include "TFIOElement.h"
#include "pil.h"

#include "widget.h"
#include "tools.h"

static Bool_t GetDol(string * hdu, string filename);

/*****************************************************************************/
/*  Function :     W_TextEntry::W_TextEntry                                  */
/*                                                                           */
/*  Description:   A text entry is creted for stirngs and filename (s and f) */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     the parameter name                  */
/*  type       input   T_STRING 1        the parameter is a string type (s)  */
/*                     T_FILE   2        the parameter is a file type (f)    */
/*                                                                           */
/*****************************************************************************/
W_TextEntry::W_TextEntry(const char * name, int type)
  : Widget(name)
{
   m_height = 2.0;
   m_width  = type == T_FILE ? g_fileW : g_textW;
   m_type   = type;
   m_fitsExt  = "-1";  // not defined
}
/*****************************************************************************/
/*  Function :     W_TextEntry::Create                                       */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_TextEntry::Create(TGCompositeFrame * win, const char * winName)
{
   if (m_z != winName)
      return;

   m_id = -1;
   CreateName(win, 0.2);

   m_entry = new TGTextEntry(win, m_buffer = 
                             new TGTextBuffer((Int_t)(m_width + 0.5)), g_id);
   m_textId = g_id;
   g_id ++;

   m_entry->SetEnabled(!m_disabled);

   if (!m_value.empty())
      m_buffer->AddText(0, m_value.c_str());

   double width = (m_type & (T_DOL | T_FILE)) ? m_width - 10 : m_width - 1;
   win->AddFrame(m_entry, new TGXYLayoutHints(
         m_xPos + m_nameW, m_yPos + 0.1, width, 1.8, 
         TGXYLayoutHints::kLRubberX | TGXYLayoutHints::kLRubberW | 
         TGXYLayoutHints::kLRubberY));

   if (m_type & (T_FILE | T_DOL))
      {
      m_button = new TGTextButton(win, "browse", g_id);
      m_id = g_id;
      g_id++;
      m_button->SetState(m_disabled ? kButtonDisabled : kButtonUp);
      win->AddFrame(m_button,  new TGXYLayoutHints(
            m_xPos + m_nameW +  m_width - 9, m_yPos + 0.1, 8, 1.8));
      }

   m_open = true;
}
/*****************************************************************************/
/*  Function :     W_TextEntry::ProcessMessage                               */
/*                                                                           */
/*  Description:   This function is called from the window to perform the    */
/*                 necessary action when the user hids this widget           */
/*                 The m_value is updated.                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  msg        input                     the ROOT message for widgets        */
/*  param1     input                     the id of the widget that was hit   */
/*                                       by the user. This function does     */
/*                                       nothing if param1 is not the id of  */
/*                                       this widget.                        */
/*  param2     input                     furter information, often not used  */
/*                                                                           */
/*****************************************************************************/
void W_TextEntry::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{
   if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == m_id)
      {

      const char * allFileTypes[]  = {"all files", "*", 0, 0};
      const char * fitsFileTypes[] = {"FITS files", "*.fits",
                                    "compressed FITS files", "*.fits.gz",
                                    "all files", "*",  0, 0,};
      TGFileInfo fi;
      if (m_type & T_DOL)
         fi.fFileTypes = fitsFileTypes; 
      else
         fi.fFileTypes = allFileTypes;

      new TGFileDialog(gClient->GetRoot(), gClient->GetRoot(),
                       kFDOpen, &fi);

      if (fi.fFilename == NULL)
         // the user pressed the cancle button
         return;

      // remove the .gz at the end
      if (strcmp(fi.fFilename + strlen(fi.fFilename) - 3, ".gz") == 0)
         fi.fFilename[strlen(fi.fFilename)-3] = 0;

      if (m_value == fi.fFilename)
         return;

      m_value = fi.fFilename;
      g_modified = kTRUE;

      
      if (m_type & T_DOL)
         {
         // it is a DOL 
         if (m_fitsExt.size() != 0)
            m_value = m_value + "[" + m_fitsExt + "]"; 	
         else
            {
            string hdu;
            if (GetDol(&hdu, m_value))
               m_value = m_value + "[" + hdu + "]";
            }
         }  

      // update the text field on the screen
      m_buffer->Clear();
      m_buffer->AddText(0, m_value.c_str());
      gClient->NeedRedraw(m_entry);
      }

   else if (GET_SUBMSG(msg) == kTE_TEXTCHANGED && param1 == m_textId)
      {
      g_modified = kTRUE;

      m_value = m_buffer->GetString(); 
      StripBlanks(m_value);
      }
}
/*****************************************************************************/
/*  Function :     W_TextEntry::ValueChanged                                 */
/*                                                                           */
/*  Description:   This function is called from other widgets if they change */
/*                 their value. This widget can be enabled or disabled       */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  widgetName  input                    parameter name of the widget which  */
/*                                       changed its value                   */
/*  value       input                    new value of widgetName             */
/*                                                                           */
/*****************************************************************************/
void W_TextEntry::ValueChanged(string & widgetName, string & value)
{
   if (widgetName != m_disableW)   return;
   m_disabled = m_disableV.find(value) != m_disableV.end();

   if (m_open)
      {
      m_entry->SetEnabled(!m_disabled);
      if (m_type & (T_FILE | T_DOL))
      m_button->SetState(m_disabled ? kButtonDisabled : kButtonUp);
      }

}
/*****************************************************************************/
/*  Function :     W_TextEntry::UpdateParamFile                              */
/*                                                                           */
/*  Description:   Writes its value into the parameter file using PIL        */
/*                 functions.                                                */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_TextEntry::UpdateParamFile()
{
   if (m_type & T_STRING)
      PILPutString(m_name.c_str(), m_value.c_str());
   else
      PILPutFname(m_name.c_str(), m_value.c_str());
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////


class TableSelect : public TGTransientFrame
{
public:
  TableSelect();
  ~TableSelect();

  void Run(TGListBox * listBox, string * hdu);

  Bool_t ProcessMessage(Long_t msg, Long_t param1, Long_t param2);
  virtual void CloseWindow();

private:

  TGListBox  * m_listBox;
  string     * m_hdu;
};
/*****************************************************************************/
/*  Function :     GetDol                                                    */
/*                                                                           */
/*  Description:   Selection of a table in a given FITS file                 */
/*                                                                           */
/*                                                                           */
/*  Return code:   kTRUE if the user selected a DOL                          */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  hdu         output                   the selected data structure         */
/*  filename    input                    the FITS file in which the user     */
/*                                       can select a table                  */
/*                                                                           */
/*****************************************************************************/
static Bool_t GetDol(string * hdu, string filename)
{
   int           status = ISDC_OK;
   int           index;
   TGListBox	  * listBox;

   // tableSelect will be deletet in its CloseWindow() - function
   TableSelect  * tableSelect = new TableSelect();
   listBox = new TGListBox(tableSelect, 1);

   TFFileIter fileIter(filename.c_str());

   index = -1;
   while (fileIter.Next())
      {
      index++;
      char name[50];
      sprintf(name, "%d: %s", 
              index, fileIter->GetName());
      listBox->AddEntry(name, index);		 	
      }

   if (index == -1)
      {
      delete listBox;
      delete tableSelect;
      new TGMsgBox(gClient->GetRoot(), gClient->GetRoot(), 
                  "Error Message", 
                  "Selected files does not has any HDU or is no FITS file!",
	                kMBIconExclamation, kMBOk);
      return kFALSE;
      }

   tableSelect->Run(listBox, hdu);

   return !hdu->empty();
}
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////


/*****************************************************************************/
/*  Function :     TableSelect::TableSelect                                  */
/*                                                                           */
/*  Description:   Constructor of a class to display a GUI with a list box   */
/*                 of all HDUs a FITS file                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*****************************************************************************/
TableSelect::TableSelect()
  : TGTransientFrame(gClient->GetRoot(), gClient->GetRoot(), 35, 14)
{
   SetLayoutManager(new TGXYLayout(this));

   SetWindowName("HDU Selection");
   SetIconName("HDU Selection");
}
/*****************************************************************************/
/*  Function :     TableSelect::Run                                          */
/*                                                                           */
/*  Description:   Show the window and wait until the user has selected      */
/*                 a table in the list box and pressed the OK button         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  listbox     input                    the listbox with all table names    */
/*  hdu         output                   the selected table name             */
/*                                                                           */
/*****************************************************************************/
void TableSelect::Run(TGListBox * listBox, string * hdu)
{
   m_listBox = listBox;
   m_hdu     = hdu;

   // create the standard buttons
   TGTextButton * button;
   button = new TGTextButton(this, "&Ok", 1);
   AddFrame(button, new TGXYLayoutHints(25, 1., 8, 1.8));
   button = new TGTextButton(this, "&Cancel", 2);
   AddFrame(button, new TGXYLayoutHints(25, 3., 8, 1.8));

   AddFrame(listBox, new TGXYLayoutHints(1.5, 1., 20, 12, 
            TGXYLayoutHints::kLRubberH));

   MapSubwindows();
   Layout();

   MapWindow();
   fClient->WaitFor(this);
}
/*****************************************************************************/
/*  Function :     TableSelect::~TableSelect                                 */
/*                                                                           */
/*  Description:   Destructor of this window. Frees memory for all child     */
/*                 frames and its layout hints                               */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
TableSelect::~TableSelect()
{
   TGFrameElement 		*ptr;

   // delete all frames and layout hints
   if (fList)
      {
      TIter next(fList);
      while ((ptr = (TGFrameElement *) next())) 
      {
      delete ptr->fLayout;
      delete ptr->fFrame;
      }
      }
}
/*****************************************************************************/
/*  Function :     TableSelect::ProcessMessage                               */
/*                                                                           */
/*  Description:   Standard ROOT function to process mouse events            */
/*                                                                           */
/*                                                                           */
/*  Return code:   always kTRUE                                              */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  See ROOT manual for the standard function parameters                     */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
Bool_t TableSelect::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{
   if (GET_SUBMSG(msg) == kCM_BUTTON)
      {
      if (param1 == 1)
         {
         Int_t sel = m_listBox->GetSelected();
         if (sel >= 0)
            {
            char str[10];
            sprintf(str, "%d", sel);
            *m_hdu = str;
            }
         }
      CloseWindow();
      }

   return kTRUE;
}
/*****************************************************************************/
/*  Function :     TableSelect::CloseWindow                                  */
/*                                                                           */
/*  Description:   Closes this window and the function Run will return.      */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void TableSelect::CloseWindow()
{
   delete this;
}



