/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*  File:      w_window.cpp                                            */
/*  Version:   1.0                                                     */
/*  Component: param_gui_lib                                           */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   An other window. It is defined with the line            */
/*             # GUI _WINDOW_ ,                                        */
/*             in the parameter file and opened by a button on its     */
/*             parent window. The parent window has the z - coordinate */
/*             m_z                                                     */
/*                                                                     */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 16/08/02 RR  First version                            */
/*                                                                     */
/***********************************************************************/

#include <TGXYLayout.h>
#include <TGButton.h>

#include "widget.h"
#include "param_gui.h"
#include "tools.h"

/*****************************************************************************/
/*  Function :     W_Window::W_Window                                        */
/*                                                                           */
/*  Description:   A window widget is a button on its m_z - window and       */
/*                 a new window with widgets are created when the user       */
/*                 press the button                                          */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name        input                    the widget name, will be always     */
/*                                       _WINDOW_                            */
/*                                                                           */
/*****************************************************************************/
W_Window::W_Window(const char * name)
  : Widget(name)
{
   m_height = 2;
   m_width  = 10;
   m_nameW  = 0;
   m_window = NULL;
   m_button = NULL;
}
/*****************************************************************************/
/*  Function :     osm_gti                                                   */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                 In the moment only the button is creted, not the window.  */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_Window::Create(TGCompositeFrame * win, const char * winName)
{
   if (m_z != winName)
      return;

   if (m_guiName.empty())
      // we need a name
      return;

   m_button = new TGTextButton(win, m_guiName.c_str(),  g_id);
   m_id = g_id;
   g_id++;

   win->AddFrame(m_button, new TGXYLayoutHints(
         m_xPos + 1, m_yPos + 0.1, m_width - 2, m_height - 0.2));

   m_open = true;
}
/*****************************************************************************/
/*  Function :     W_Window::ProcessMessage                                  */
/*                                                                           */
/*  Description:   This function is called from the window to perform the    */
/*                 necessary action when the user hids this widget:          */
/*                 The associated new window is created                      */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  msg        input                     the ROOT message for widgets        */
/*  param1     input                     the id of the widget that was hit   */
/*                                       by the user. This function does     */
/*                                       nothing if param1 is not the id of  */
/*                                       this widget.                        */
/*  param2     input                     furter information, often not used  */
/*                                                                           */
/*****************************************************************************/
void W_Window::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{
   if (param1 != m_id || GET_SUBMSG(msg) != kCM_BUTTON)
      return;

   m_window = new SubWindow(m_guiName);
 
   m_button->SetState(kButtonDisabled);

//   gClient->WaitFor(m_window);
}
/*****************************************************************************/
/*  Function :     W_Window::CloseWidget                                     */
/*                                                                           */
/*  Description:   This function is called when the window in which the      */
/*                 button is created is closed. The associated window of     */
/*                 this button will be closed if it is open.                 */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  backGround  input                    the name of the window that is      */
/*                                       closed. The name of the window      */
/*                                       is the z - coordinate of that       */
/*                                       window and the text of its button   */
/*                                                                           */
/*****************************************************************************/
void W_Window::CloseWidget(string & backGround) 
{
   Widget::CloseWidget(backGround);

   if (backGround == m_guiName)
      {
      m_button->SetState(kButtonUp);
      m_window = NULL;
      }

   if (backGround == m_z && m_window)
      m_window->CloseWindow();
      
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     SubWindow::SubWindow                                      */
/*                                                                           */
/*  Description:   This is a new window on which further windgets can be     */
/*                 created. It has its own unique z - coordinate, which      */
/*                 is the m_guiName of the associated W_Window class and the */
/*                 text of the accociated button.                            */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name        input                    name and z - coordinate of this     */
/*                                       window                              */
/*                                                                           */
/*****************************************************************************/
SubWindow::SubWindow(const string & name)
  : TGMainFrame(gClient->GetRoot(), 200, 50)
{	
   m_name    = name;

   StdButton stdButton(1);
   WinSize   winSize(stdButton);

   // set the window size depending of the defined size of
   // the widgets
   I_Widgets end = PG_MainWindow::EndW();
   I_Widgets begin = PG_MainWindow::BeginW();
   for (I_Widgets i_w = begin; i_w != end; i_w++)
      winSize.SetSize(*i_w, m_name.c_str());

   // set the positon of the other widgets and than set the
   // window size depending of the size of these widgets
   winSize.SetMode(1);
   for (I_Widgets i_w = begin; i_w != end; i_w++)
      winSize.SetSize(*i_w, m_name.c_str());

   Resize((UInt_t)(winSize.GetWidth() + 1.0), 
          (UInt_t)(winSize.GetHeight() + 1.0));

   SetLayoutManager(new TGXYLayout(this));

   // draw all widgets, first the frames
   for (I_Widgets i_w = begin; i_w != end; i_w++)
      (*i_w)->CreateFrame(this, m_name.c_str());
   for (I_Widgets i_w = begin; i_w != end; i_w++)
      (*i_w)->Create(this, m_name.c_str());

   // draw the standard buttons
   stdButton.Create(this, &winSize);

   MapSubwindows();
   Layout();
   SetWindowName(m_name.c_str());
   MapWindow();

   for (I_Widgets i_w = begin; i_w != end; i_w++)
      (*i_w)->BeforeOpen(m_name);

}
/*****************************************************************************/
/*  Function :     SubWindow:: ~SubWindow                                    */
/*                                                                           */
/*  Description:   Deletes all widgets on this window and their layout       */
/*                 managers                                                  */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
SubWindow:: ~SubWindow()
{
   TGFrameElement *ptr;

   // delete all frames and layout hints
   if (fList)
      {
      TIter next(fList);
      while ((ptr = (TGFrameElement *) next())) 
         {
         delete ptr->fLayout;
         delete ptr->fFrame;
         }
   }
}
/*****************************************************************************/
/*  Function :     SubWindow::ProcessMessage                                 */
/*                                                                           */
/*  Description:   The user pressed the mouse on this window. Either         */
/*                 a standard button was hit or one of the widgets on this   */
/*                 window. In second case the message is forwarded to        */
/*                 the widget.                                               */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  See ROOT user manual                                                     */
/*                                                                           */
/*****************************************************************************/
Bool_t SubWindow::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{

   if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == BT_OK)
      CloseWindow();

   else if (GET_SUBMSG(msg) == kCM_BUTTON && param1 == BT_HELP)
      OpenHelp(this, PG_MainWindow::GetParamFileName());

   else
      for (I_Widgets i_w = PG_MainWindow::BeginW(); 
            i_w != PG_MainWindow::EndW(); i_w++)
         (*i_w)->ProcessMessage(msg, param1, param2);

   return kTRUE;
}
/*****************************************************************************/
/*  Function :     SubWindow::CloseWindow                                    */
/*                                                                           */
/*  Description:   Closed this window and informs all widget that this       */
/*                 window is closed.                                         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void SubWindow::CloseWindow()
{
   for (I_Widgets i_w = PG_MainWindow::BeginW(); 
         i_w != PG_MainWindow::EndW(); i_w++)
      (*i_w)->CloseWidget(m_name);

   delete this;
}
    
