/***********************************************************************/
/*                                                                     */
/*                    INTEGRAL SCIENCE DATA CENTRE                     */
/*                                                                     */
/*                      PARAMETER - GUI LIBRARY                        */
/*                                                                     */
/*                                                                     */
/*  File:      widget.cpp                                              */
/*  Version:   1.0                                                     */
/*  Component: param_gui_lib                                           */
/*                                                                     */
/*  Authors:   Reiner Rohlfs (RR)                                      */
/*             ISDC                                                    */
/*                                                                     */
/*  Purpose:   Here the base class Widget and some smaller derived     */
/*             classes like W_CheckButton, W_IntEntry, W_Frame, amd    */
/*             W_RealEntry are coded.                                  */
/*                                                                     */
/*  Revision History:                                                  */
/*                                                                     */
/*  version 1.0: 15/08/02 RR  First version                            */
/*                                                                     */
/***********************************************************************/
#include <stdio.h>
#include <ctype.h>

#include "pil.h"

#include <TGLabel.h>
#include <TGWidget.h>
#include <TGNumberEntry.h>
#include <TGXYLayout.h>

#include "param_gui.h"
#include "tools.h"
#include "widget.h"

Int_t  Widget::g_id = 100;    // the id of the widgets

Bool_t Widget::g_modified = kFALSE;

double Widget::g_textW = 31;  // default width of string entry
                              // does include 1 space
double Widget::g_fileW = 81;  // default width of file entry
                              // does include 10 space for button

double Widget::g_nameW = 16;  // default width of paramter name
                              // does include 2 spaces


/*****************************************************************************/
/*  Function :     Widget::Widget                                            */
/*                                                                           */
/*  Description:   Constructor of the base class of all widgets. Here the    */
/*                 parameter name is set which cannot be changed             */
/*                 afterwards.                                               */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     parameter name of this widget and   */
/*                                       its assosiated parameter            */
/*                                                                           */
/*****************************************************************************/
Widget::Widget(const char * name)
  : m_name(name)
{
   m_xPos     = 
   m_yPos     = -1;   // undefined
   m_nameW    = g_nameW;
   m_width    = 0;
   m_height   = 2;
   m_open     = false;
   m_z        = "main";
   m_disabled = false;
   m_type     = 0;
}
/*****************************************************************************/
/*  Function :     Widget::SetMode                                           */
/*                                                                           */
/*  Description:   Set the parameter mode (q, h, ql ...)                     */
/*                 Hidden parameters are place on the "hidden" window by     */
/*                 default                                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  mode        input                    the mode of the parameter           */
/*                                                                           */
/*****************************************************************************/
void Widget::SetMode(const char * mode)
{
   m_mode      = mode;

   if (strchr(mode, 'h'))
      m_z = "hidden";
}
/*****************************************************************************/
/*  Function :     Widget::SetAttribute                                      */
/*                                                                           */
/*  Description:   Set a new attribute to this widget                        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  attrStr    input                     the text between two commas ,xxx,   */
/*                                       in the "# GUI" comment line         */
/*                                       for this widget.                    */
/*                                                                           */
/*****************************************************************************/
void Widget::SetAttribute(const char * attrStr)
{
   if (strncmp(attrStr, "X:", 2) == 0)
      sscanf(attrStr+2, "%lf", &m_xPos);


   else if (strncmp(attrStr, "Y:", 2) == 0)
      sscanf(attrStr+2, "%lf", &m_yPos);

   else if (strncmp(attrStr, "Z:", 2) == 0)
      {
      m_z = attrStr + 2;
      StripBlanks(m_z);
      }

  // not for folders
   else if (strncmp(attrStr, "W:", 2) == 0 && 
            strncmp(m_name.c_str(), "_FOLDER_", 8) != 0 )
      sscanf(attrStr+2, "%lf", &m_width);

  // only for frames
   else if (strncmp(attrStr, "H:", 2) == 0 &&
            strncmp(m_name.c_str(), "_FRAME_", 7) == 0 )
      sscanf(attrStr+2, "%lf", &m_height);

   else if (strncmp(attrStr, "TABS:", 5) == 0)
      m_tabNames = attrStr + 5;

   else if (strncmp(attrStr, "NAME_W:", 7) == 0)
      sscanf(attrStr+7, "%lf", &m_nameW);

   else if (strncmp(attrStr, "NAME:", 5) == 0)
      {
      m_guiName = attrStr+5;
      StripBlanks(m_guiName);
      }

   else if (strncmp(attrStr, "DOL:", 4) == 0)
      {
      m_type |= T_DOL;
      m_fitsExt = attrStr + 4;
      StripBlanks(m_fitsExt);
      }

   else if (strncmp(attrStr, "DISABLE:", 8) == 0)
      {
      char tmp[500];
      strcpy(tmp, attrStr);
      char * pos = strchr(tmp + 8, '=');
      if (pos == NULL) return;
      
      *pos = 0;
      string widget = tmp + 8;
      StripBlanks(widget);
      if (m_disableW != widget)
         m_disableV.clear();
      m_disableW = widget;

      SubStr sub(pos + 1);
      string str;
      while (sub.Next(str))
         m_disableV.insert(str);
      }
}
/*****************************************************************************/
/*  Function :     Widget::SetCommonAttribute                                */
/*                                                                           */
/*  Description:   Set a common attribute for all widgets                    */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  attrStr     input                    the attribute   from the parameter  */
/*                                       file in the line which started      */
/*                                       # GUI _COMMON_,                     */
/*                                                                           */
/*****************************************************************************/
void Widget::SetCommonAttribute(const char * attrStr)
{
   if (strncmp(attrStr, "NAME_W:", 7) == 0)
      sscanf(attrStr+7, "%lf", &g_nameW);

   if (strncmp(attrStr, "TEXT_W:", 7) == 0)
      sscanf(attrStr+7, "%lf", &g_textW);

   if (strncmp(attrStr, "FILE_W:", 7) == 0)
      sscanf(attrStr+7, "%lf", &g_fileW);

}
/*****************************************************************************/
/*  Function :     Widget::CreateName                                        */
/*                                                                           */
/*  Description:   Create the name TGLabel to display the parameter name     */
/*                 and a tooltip with the prompt string                      */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     the name label will be placed on    */
/*                                       this window                         */
/*  yOff       input                     the offset in Y direction to place  */
/*                                       the label. Default value: 0.2       */
/*                                                                           */
/*****************************************************************************/
void Widget::CreateName(TGCompositeFrame * win, double yOff)
{
   char str[50];

   if (m_guiName.empty())
      strncpy(str, m_name.c_str(), 49);
   else
      strncpy(str, m_guiName.c_str(), 49);

   str[49] = 0;
   strcat(str, ":"); 

   ToolTipLabel * label = new ToolTipLabel(win, str);
   label->SetTextJustify(kTextRight);

   if (!m_prompt.empty())
      label->SetToolTip(m_prompt.c_str());

   win->AddFrame(label, new TGXYLayoutHints(
                 m_xPos + 1, m_yPos + yOff, m_nameW - 2, 1.6));

}

/*****************************************************************************/
/*  Function :     Widget::GetRunComment                                     */
/*                                                                           */
/*  Description:   Prepares the command line string for this widget:         */
/*                 (paramName=value ) if value is not empty                  */
/*                                                                           */
/*  Return code:   paramName=value                                           */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*****************************************************************************/
std::string Widget::GetRunComment()
{
   if (m_value.length() > 0)
      return m_name + "=" + m_value + " ";

   return m_value;
}
/*****************************************************************************/
/*  Function :     Widget::WriteParamFile                                    */
/*                                                                           */
/*  Description:   Writes one line into a new parameter line, describing     */
/*                 this widget.                                              */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  fl          input                    file handler                        */
/*                                                                           */
/*****************************************************************************/
void Widget::WriteParamFile(FILE * fl)
{
   if (m_paramType.empty())
      // this is a widget without parameter
      return;

   fprintf(fl, "%s,%s,%s,",
           m_name.c_str(), m_paramType.c_str(),
           m_mode.c_str() );

   if (m_paramType.at(0) == 's' || m_paramType.at(0) == 'f')      
      fprintf(fl, "\"%s\"", m_value.c_str() );
   else
      fprintf(fl, "%s", m_value.c_str() );
           
   fprintf(fl, ",%s,%s,\"%s\"\n",
           m_min.c_str(), m_max.c_str(), m_prompt.c_str() );
}
/*****************************************************************************/
/*  Function :     Widget::WriteWidgetAttr                                   */
/*                                                                           */
/*  Description:   Writes one line into a new parameter line, describing     */
/*                 the attributes of this widget.                            */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  fl          input                    file handler                        */
/*                                                                           */
/*****************************************************************************/
void Widget::WriteWidgetAttr(FILE * fl)
{
   if (m_name == "_HIDDEN_")
      return;

   // GUI identifier and name
   fprintf(fl, "# GUI %s", m_name.c_str() );

   // X, Y, Z coordinates
   fprintf(fl, ", X:%6.2f, Y:%6.2f, Z:%s", m_xPos, m_yPos, m_z.c_str());

   // W  (width)
   if ( !((m_type & T_STRING) && m_width == g_textW) &&
        !((m_type & T_FILE)   && m_width == g_fileW) &&
        strncmp("_FOLDER_", m_name.c_str(), 8) != 0     )
      fprintf(fl, ", W:%6.2f", m_width);

   // H height only for frames
   if (strncmp("_FRAME_", m_name.c_str(), 7) == 0)
      fprintf(fl, ", H:%6.2f", m_height);

   // Name attribute
   if (!m_guiName.empty() &&
       strncmp("_FOLDER_", m_name.c_str(), 8) != 0 )
      fprintf(fl, ", NAME:%s", m_guiName.c_str());

   // name width only for entry widgets
   if (!m_paramType.empty() &&
       m_nameW != g_nameW          )
      fprintf(fl, ", NAME_W:%6.2f", m_nameW);

   // tab names for folder widgets
   if (strncmp("_FOLDER_", m_name.c_str(), 8) == 0 &&
       !m_tabNames.empty() )
      fprintf(fl, ", TABS:%s", m_tabNames.c_str() );

   // DOL
   if (m_type & T_DOL)
      fprintf(fl, ", DOL:%s", m_fitsExt.c_str());

   // disable (only for entry widgets)
   if (!m_paramType.empty() &&
       !m_disableW.empty()     )
      {
      // write it in a new line
      fprintf(fl, "\n# GUI %s, DISABLE: %s =",m_name.c_str(), m_disableW.c_str() );
      
      std::set<std::string>::const_iterator i_name = m_disableV.begin();
      while (i_name != m_disableV.end() )
         {
         if (i_name != m_disableV.begin() )
            fprintf(fl, " |");

         fprintf(fl, " %s", i_name->c_str() );
         i_name++;
         }
      }


   fprintf(fl, "\n");
}
/*****************************************************************************/
/*  Function :     Widget::WriteCommonAttr                                   */
/*                                                                           */
/*  Description:   Writes one line into a new parameter line, describing     */
/*                 the common attributes. This is a static function.         */
/*                                                                           */
/*  Return code:   none                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  fl          input                    file handler                        */
/*                                                                           */
/*****************************************************************************/
void Widget::WriteCommonAttr(FILE * fl)
{
   fprintf(fl, "# GUI _COMMON_, NAME_W: %6.2f, TEXT_W: %6.2f, FILE_W: %6.2f\n#\n",
           g_nameW, g_textW, g_fileW);
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     ToolTipLabel::ToolTipLabel                                */
/*                                                                           */
/*  Description:   A TGLabel with a tooltip. It is used to show the          */
/*                 prompt sting of a parameter while the mouse moves above   */
/*                 the parameter name                                        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win         input                    the label will be place on this     */
/*                                       window                              */
/*  text        input                    text of the lable. Either the       */
/*                                       parameter name or the name defined  */
/*                                       by the attribute NAME:              */
/*                                                                           */
/*****************************************************************************/
ToolTipLabel::ToolTipLabel(const TGCompositeFrame * win, const char * text)
    : TGLabel(win, text)  
{ 
   m_tip = NULL;

   AddInput(kEnterWindowMask | kLeaveWindowMask);
}
/*****************************************************************************/
/*  Function :     ToolTipLabel::SetToolTip                                  */
/*                                                                           */
/*  Description:   Set the tooltip text to this label                        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  tip         input                    tool tip text                       */
/*                                                                           */
/*****************************************************************************/
void ToolTipLabel::SetToolTip(const char * tip)
{
   m_tip = new TGToolTip(fClient->GetRoot(), this, tip, 500);
}
/*****************************************************************************/
/*  Function :     ToolTipLabel::HandleCrossing                              */
/*                                                                           */
/*  Description:   Handler function to display and to remove the tooltip     */
/*                                                                           */
/*  Return code:   kTRUE                                                     */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  event      input                     see ROOT manual                     */
/*                                                                           */
/*****************************************************************************/
Bool_t ToolTipLabel::HandleCrossing(Event_t *event) 
{ 
   if (m_tip) 
      {
      if (event->fType == kEnterNotify)
         m_tip->Reset();
      else
         m_tip->Hide();
   }
   return kTRUE;
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     W_CheckButton::W_CheckButton                              */
/*                                                                           */
/*  Description:   A check - button widget for boolean parameters (b)        */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     the parameter name                  */
/*                                                                           */
/*****************************************************************************/
W_CheckButton::W_CheckButton(const char * name)
  : Widget(name)
{
   m_height = 2.0;
   m_width  = 16;
}
/*****************************************************************************/
/*  Function :     W_CheckButton::Init                                       */
/*                                                                           */
/*  Description:   Initialize all other widgets (enable, disable them)       */
/*                 depending on the value of this widget. This is called     */
/*                 just after the widget classes with all attributes were    */
/*                 created, but before the screen widgets were created.      */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_CheckButton::Init()
{
   if (m_value.empty())
      return;

   if ( tolower(m_value.at(0)) == 'y')
      m_value = "yes";
   else if (tolower(m_value.at(0)) == 'n')
      m_value = "no";

   for (I_Widgets i_w = PG_MainWindow::BeginW(); 
         i_w != PG_MainWindow::EndW(); i_w++)
      (*i_w)->ValueChanged(m_name, m_value);
}
/*****************************************************************************/
/*  Function :     W_CheckButton::Create                                     */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_CheckButton::Create(TGCompositeFrame * win, const char * winName)
{
   if (m_z != winName)
      return;

   CreateName(win, 0.2);

   m_button = new TGCheckButton(win, "checked: yes", g_id);
   m_id = g_id;
   g_id++;

   if (!m_value.empty() && m_value.at(0) == 'y')
      m_button->SetState(kButtonDown);

   if (m_disabled)
      m_button->SetState(kButtonDisabled);

   win->AddFrame(m_button, new TGXYLayoutHints(
         m_xPos + m_nameW, m_yPos + 0.2, m_width - 1, 1.6));

   m_open = true;
}
/*****************************************************************************/
/*  Function :     W_CheckButton::ProcessMessage                             */
/*                                                                           */
/*  Description:   This function is called from the window to perform the    */
/*                 necessary action when the user hids this widget           */
/*                 The m_value is updated.                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  msg        input                     the ROOT message for widgets        */
/*  param1     input                     the id of the widget that was hit   */
/*                                       by the user. This function does     */
/*                                       nothing if param1 is not the id of  */
/*                                       this widget.                        */
/*  param2     input                     furter information, often not used  */
/*                                                                           */
/*****************************************************************************/
void W_CheckButton::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{
   if (param1 != m_id || GET_SUBMSG(msg) != kCM_CHECKBUTTON)
      return;


   if (m_button->GetState() == kButtonDown)
      {
      if (m_value == "yes")  return;
      m_value = "yes";
      }
   else if (m_button->GetState() == kButtonUp)
      {
      if (m_value == "no")   return;
      m_value = "no";
      }
   else
      return;

   g_modified = kTRUE;


   // inform other widget that this widget has changed its value
   // they may be disabled or enabled
   for (I_Widgets i_w = PG_MainWindow::BeginW();
         i_w != PG_MainWindow::EndW(); i_w++)
      (*i_w)->ValueChanged(m_name, m_value);
}
/*****************************************************************************/
/*  Function :     W_CheckButton::ValueChanged                               */
/*                                                                           */
/*  Description:   This function is called from other widgets if they change */
/*                 their value. This widget can be enabled or disabled       */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  widgetName  input                    parameter name of the widget which  */
/*                                       changed its value                   */
/*  value       input                    new value of widgetName             */
/*                                                                           */
/*****************************************************************************/
void W_CheckButton::ValueChanged(string & widgetName, string & value)
{
   if (widgetName != m_disableW)   return;
   m_disabled = m_disableV.find(value) != m_disableV.end();

   if (m_open)
      {
      EButtonState state;
      if (!m_value.empty() && m_value.at(0) == 'y')
         state = kButtonDown;
      else
         state = kButtonUp;
      m_button->SetState(m_disabled ? kButtonDisabled : state);
      }
}
/*****************************************************************************/
/*  Function :     W_CheckButton::UpdateParamFile                            */
/*                                                                           */
/*  Description:   Writes its value into the parameter file using PIL        */
/*                 functions.                                                */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_CheckButton::UpdateParamFile()
{
   if (!m_value.empty() && m_value.at(0) == 'y')
      PILPutBool(m_name.c_str(), 1);
   else
      PILPutBool(m_name.c_str(), 0);
}
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     W_IntEntry::W_IntEntry                                    */
/*                                                                           */
/*  Description:   A integer entry is created for int parameters (i)         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     the parameter name                  */
/*                                                                           */
/*****************************************************************************/
W_IntEntry::W_IntEntry(const char * name)
  : Widget(name)
{
   m_height = 2.0;
   m_width  = 14;
}
/*****************************************************************************/
/*  Function :     W_IntEntry::Create                                        */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_IntEntry::Create(TGCompositeFrame * win, const char * winName)
{
   if (m_z != winName)
      return;

   CreateName(win, 0.2);

   m_entry = new TGNumberEntry(win, 0, 5, g_id);
   m_id = g_id;
   g_id++;

   m_entry->SetFormat(TGNumberFormat::kNESInteger);

   if (!m_value.empty())
      {
      Long_t val = atoi(m_value.c_str());
      m_entry->SetIntNumber(val);
      }

   TGNumberFormat::ELimit limit = TGNumberFormat::kNELNoLimits;
   Double_t min, max;
   if (!m_min.empty())
      {
      min = atoi(m_min.c_str());
      limit = TGNumberFormat::kNELLimitMin;
      }

   if (!m_max.empty())
      {
      max = atoi(m_max.c_str());
      limit = (TGNumberFormat::ELimit)(limit | TGNumberFormat::kNELLimitMax);
      }
   m_entry->SetLimits(limit, min, max);

   m_entry->SetState(!m_disabled);
   win->AddFrame(m_entry, new TGXYLayoutHints(
         m_xPos + m_nameW, m_yPos + 0.2, m_width - 1, 1.6,
         TGXYLayoutHints::kLRubberX | TGXYLayoutHints::kLRubberW | 
         TGXYLayoutHints::kLRubberY));

   m_open = true;
}
/*****************************************************************************/
/*  Function :     W_IntEntry::ProcessMessage                                */
/*                                                                           */
/*  Description:   This function is called from the window to perform the    */
/*                 necessary action when the user hids this widget           */
/*                 The m_value is updated.                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  msg        input                     the ROOT message for widgets        */
/*  param1     input                     the id of the widget that was hit   */
/*                                       by the user. This function does     */
/*                                       nothing if param1 is not the id of  */
/*                                       this widget.                        */
/*  param2     input                     furter information, often not used  */
/*                                                                           */
/*****************************************************************************/
void W_IntEntry::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{
   if (param1 != m_id || GET_SUBMSG(msg) != kTE_TEXTCHANGED)
      return;

   Long_t val = m_entry->GetIntNumber();

   char valStr[30];
   sprintf(valStr, "%d", val);
   if (m_value == valStr)
      return;

   g_modified = kTRUE;
   m_value = valStr;
}
/*****************************************************************************/
/*  Function :     W_IntEntry::ValueChanged                                  */
/*                                                                           */
/*  Description:   This function is called from other widgets if they change */
/*                 their value. This widget can be enabled or disabled       */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  widgetName  input                    parameter name of the widget which  */
/*                                       changed its value                   */
/*  value       input                    new value of widgetName             */
/*                                                                           */
/*****************************************************************************/
void W_IntEntry::ValueChanged(string & widgetName, string & value)
{
   if (widgetName != m_disableW)   return;
   m_disabled = m_disableV.find(value) != m_disableV.end();

   if (m_open)
      m_entry->SetState(!m_disabled);
}
/*****************************************************************************/
/*  Function :     W_IntEntry::UpdateParamFile                               */
/*                                                                           */
/*  Description:   Writes its value into the parameter file using PIL        */
/*                 functions.                                                */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_IntEntry::UpdateParamFile()
{
   if (!m_value.empty())
      PILPutInt(m_name.c_str(), atoi(m_value.c_str()) );
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     W_RealEntry::W_RealEntry                                  */
/*                                                                           */
/*  Description:   A real entry is created for real parameters (r)           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     the parameter name                  */
/*                                                                           */
/*****************************************************************************/
W_RealEntry::W_RealEntry(const char * name)
  : Widget(name)
{
   m_height = 2.0;
   m_width  = 14;
}
/*****************************************************************************/
/*  Function :     W_RealEntry::Create                                       */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_RealEntry::Create(TGCompositeFrame * win, const char * winName)
{
   if (m_z != winName)
      return;

   CreateName(win, 0.2);

   m_entry = new TGNumberEntry(win, 0, 5, g_id);
   m_id = g_id;
   g_id++;

   if (!m_value.empty())
      {
      Double_t val;
      sscanf(m_value.c_str(), "%lf", &val);
      m_entry->SetNumber(val);
      }

   TGNumberFormat::ELimit limit = TGNumberFormat::kNELNoLimits;
   Double_t min, max;
   if (!m_min.empty())
      {
      sscanf(m_min.c_str(), "%lf", &min);
      limit = TGNumberFormat::kNELLimitMin;
      }

   if (!m_max.empty())
      {
      sscanf(m_max.c_str(), "%lf", &max);
      limit = (TGNumberFormat::ELimit)
               (limit | TGNumberFormat::kNELLimitMax);
      }
   m_entry->SetLimits(limit, min, max);

   m_entry->SetState(!m_disabled);
   win->AddFrame(m_entry, new TGXYLayoutHints(
         m_xPos + m_nameW, m_yPos + 0.2, m_width - 1, 1.6,
         TGXYLayoutHints::kLRubberX | TGXYLayoutHints::kLRubberW | 
         TGXYLayoutHints::kLRubberY));

   m_open = true;
}
/*****************************************************************************/
/*  Function :     W_RealEntry::ProcessMessage                               */
/*                                                                           */
/*  Description:   This function is called from the window to perform the    */
/*                 necessary action when the user hids this widget           */
/*                 The m_value is updated.                                   */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  msg        input                     the ROOT message for widgets        */
/*  param1     input                     the id of the widget that was hit   */
/*                                       by the user. This function does     */
/*                                       nothing if param1 is not the id of  */
/*                                       this widget.                        */
/*  param2     input                     furter information, often not used  */
/*                                                                           */
/*****************************************************************************/
void W_RealEntry::ProcessMessage(Long_t msg, Long_t param1, Long_t param2)
{
   if (param1 != m_id || GET_SUBMSG(msg) != kTE_TEXTCHANGED)
      return;

   Double_t val = m_entry->GetNumber();

   char valStr[30];
   sprintf(valStr, "%g", val);
   if (m_value == valStr)
      return;

   g_modified = kTRUE;
   m_value = valStr;
}
/*****************************************************************************/
/*  Function :     W_RealEntry::ValueChanged                                 */
/*                                                                           */
/*  Description:   This function is called from other widgets if they change */
/*                 their value. This widget can be enabled or disabled       */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  widgetName  input                    parameter name of the widget which  */
/*                                       changed its value                   */
/*  value       input                    new value of widgetName             */
/*                                                                           */
/*****************************************************************************/
void W_RealEntry::ValueChanged(string & widgetName, string & value)
{
   if (widgetName != m_disableW)   return;
   m_disabled = m_disableV.find(value) != m_disableV.end();

   if (m_open)
      m_entry->SetState(!m_disabled);
      
}
/*****************************************************************************/
/*  Function :     W_RealEntry::UpdateParamFile                              */
/*                                                                           */
/*  Description:   Writes its value into the parameter file using PIL        */
/*                 functions.                                                */
/*                                                                           */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/
void W_RealEntry::UpdateParamFile()
{
   if (!m_value.empty())
      {
      double val;
      sscanf(m_value.c_str(), "%lf", &val);
      PILPutReal(m_name.c_str(), val );
      }
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*****************************************************************************/
/*  Function :     W_Frame::W_Frame                                          */
/*                                                                           */
/*  Description:   A frame with an optional title                            */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name       input                     the parameter name                  */
/*                                                                           */
/*****************************************************************************/
W_Frame::W_Frame(const char * name)
  : Widget(name)
{
   m_height = 6.0;
   m_width  = 16;
   m_nameW  = 0;
}
/*****************************************************************************/
/*  Function :     W_Frame::CreateFrame                                      */
/*                                                                           */
/*  Description:   Creates this widget in the window win, but only if this   */
/*                 widget should be inside the window called winName         */
/*                                                                           */
/*  Return code:   None                                                      */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  win        input                     this widget is created on this      */
/*                                       window                              */
/*  winName    input                     the name ( z - coordinate) of win.  */
/*                                       This widget is not created if it    */
/*                                       has to be in an other window than   */
/*                                       winName                             */
/*                                                                           */
/*****************************************************************************/
void W_Frame::CreateFrame(TGCompositeFrame * win, const char * winName)
{
   if (m_z != winName)
      return;

   TGGroupFrame * frame = new TGGroupFrame(win, "");

   if (!m_guiName.empty())
      frame->SetTitle(m_guiName.c_str());

   win->AddFrame(frame, new TGXYLayoutHints(
         m_xPos + 0.5, m_yPos, m_width - 1, m_height - 0.25,
         TGXYLayoutHints::kLRubberX | TGXYLayoutHints::kLRubberY | 
         TGXYLayoutHints::kLRubberW | TGXYLayoutHints::kLRubberH));

   m_open = true;
}

