// ///////////////////////////////////////////////////////////////////
//
//  File:      tfconvert.cxx
//
//  Version:   1.0
//
//  Author:    Reiner Rohlfs
//
//  History:
//
// ///////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "pil.h"
#include "param_api.h"

#include "TFError.h"
#include "TFGroup.h"
#include "TFImage.h"


#ifdef WIN32
#define  DOT_SLASH   ".\\"
#define  DS          '\\'       // Directory Seperator 
#define  DS2         "\\"       // Directory Seperator 

#else 
#define  DOT_SLASH   "./"
#define  DS          '/'        // Directory Seperator 
#define  DS2         "/"        // Directory Seperator 

#endif

#define L_PATH       1024

struct Params
{
   char file[PIL_LINESIZE];
   char fileFormat[PIL_LINESIZE];
   char name[PIL_LINESIZE];
   int  cycle;
   int  group;
   int  sameDir;
   char outDir[PIL_LINESIZE];
   int  compLevel;
   int  logLevel;
};

struct Count
{
   Count() {conv = 0; groups = 0;}
   int conv;
   int groups;
};

static void         ConvertGroup(TFGroup * group, const 
                          TFFilePath & inGroupName,
                          const TFFilePath & outGroupName, Params & params,
                          Count & count);
static const char * GetOutFile(Params & params, const char * inFile);
static const char * ConvertExtension(const char * fileName, Params & params);
static bool         ElementExist(const char * fileName, TFIOElement * element, 
                                 Int_t * cycle);
static char *       strrstr(char * str, const char * sub);
static void         MakeDir(const char * inName);
static void         MakeDirRecursive(char * name);


//_____________________________________________________________________________
static int ReadParams(Params * params, int argc, char ** argv)
{
//  reads all program parameters either from the command line or from the
//  parameter file. PIL is used.
   int status = ISDC_OK;
   
   for (int par = 1; par < argc; par++)
      if (strcmp("--g",   argv[par]) == 0 ||
          strcmp("--gui", argv[par]) == 0    )
         {
         // use the GUI to ask for the parameters
         ParamGui gui(argc, argv);
         if (gui.Quit())        exit(0);
         if (gui.Error() != ISDC_OK)   return gui.Error();
      
         strcpy(params->file,       gui.GetString("file"));
         strcpy(params->fileFormat, gui.GetString("outFileType"));
         strcpy(params->name,       gui.GetString("name"));
         strcpy(params->outDir,     gui.GetString("outDir"));
         params->cycle =     gui.GetInt("cycle");
         params->group =     gui.GetBool("group");
         params->sameDir =   gui.GetBool("sameDir");
         params->compLevel = gui.GetInt("compLevel");
         params->logLevel =  gui.GetInt("logLevel");

         for (int ch = 0; ch < 4; ch++)
            params->fileFormat[ch] = tolower(params->fileFormat[ch]);

         return gui.Error();
         }


   do
      {
      status = PILInit(argc, argv);
      if (status != ISDC_OK) break;

      status = PILGetFname("file", params->file);
      if (status != ISDC_OK) break;

      status = PILGetString("outFileType", params->fileFormat);
      if (status != ISDC_OK) break;
      for (int ch = 0; ch < 4; ch++)
         params->fileFormat[ch] = tolower(params->fileFormat[ch]);

      status = PILGetString("name", params->name);
      if (status != ISDC_OK) break;

      if (params->name[0] != 0)
         {
         status = PILGetInt("cycle", &params->cycle);
         if (status != ISDC_OK) break;
         }
      
      status = PILGetBool("group", &params->group);
      if (status != ISDC_OK) break;
      
      status = PILGetBool("sameDir", &params->sameDir);
      if (status != ISDC_OK) break;

      if (!params->sameDir)
         {
         status = PILGetString("outDir", params->outDir);
         if (status != ISDC_OK) break;
         }

      status = PILGetInt("compLevel", &params->compLevel);
      if (status != ISDC_OK) break;

      status = PILGetInt("logLevel", &params->logLevel);
      if (status != ISDC_OK) break;


      } while (0);

   PILClose(ISDC_OK);

   return status;
}
//_____________________________________________________________________________
main(int argc, char * argv[])
{
   Params params;
   Count  count;

   int status = ReadParams(&params, argc, argv);
   if (status != ISDC_OK)
      {
      TFError::SetError("main", "Got error %d while reading parameters", status); 
      TFError::PrintErrors();
      return 0;
      }


   TFError::SetErrorType(kAllErr);
   try{

   if (params.name[0] == 0)
      {
      // convert all elements in file
      TFFileIter fl(params.file);

      while (fl.Next())
         {
         TFFilePath inAbsName(fl->GetFileName());
         inAbsName.MakeAbsolutePath(DOT_SLASH);

         TFFilePath outAbsName(GetOutFile(params, fl->GetFileName() ));
         outAbsName.MakeAbsolutePath(DOT_SLASH);

         if (params.group && fl->IsA() == TFGroup::Class())
            ConvertGroup((TFGroup*)&*fl, inAbsName, outAbsName, 
                          params, count);

         // save the element in the new file
         try 
            {
            const char * newFileName = ConvertExtension(outAbsName.Data(), params);
            Int_t cycle;
            if (!ElementExist(newFileName, &*fl, &cycle))
               {
               MakeDir(newFileName);
               fl->SaveElement(newFileName, params.compLevel);
               count.conv++;
               }
            }
         catch (TFException) {}
         }
      }
   else
      {
      // only one element
      TFIOElement * element = TFRead(params.file, params.name, params.cycle);
      if (element)
         {

         TFFilePath inAbsName(element->GetFileName());
         inAbsName.MakeAbsolutePath(DOT_SLASH);

         TFFilePath outAbsName(GetOutFile(params, element->GetFileName() ));
         outAbsName.MakeAbsolutePath(DOT_SLASH);

         if (params.group && element->IsA() == TFGroup::Class())
            ConvertGroup((TFGroup*)element, inAbsName, outAbsName,
                          params, count);

         // save the element in the new file
         try
            {
            MakeDir(ConvertExtension(outAbsName.Data(), params));
            element->SaveElement(ConvertExtension(outAbsName.Data(), params), 
                                 params.compLevel);
            count.conv++;
            }
         catch (TFException) {}

         delete element;
         }
      }

   if (params.logLevel & 3)
      printf("converted %d data structures and %d groups\n", 
             count.conv, count.groups);
   }
   catch (TFException) {}
   TFError::PrintErrors();
}
//_____________________________________________________________________________
static void  ConvertGroup(TFGroup * group, const TFFilePath & inGroupName, 
                          const TFFilePath & outGroupName, Params & params,
                          Count & count)
{
//   converts all data structure of a group. 
   if (params.logLevel & 1)
      printf("converting group %s\n", inGroupName.Data());  

   try {
      TFGroupCol & col = (TFGroupCol&)group->GetColumn(GROUP_COL_NAME);

      UInt_t numRows = col.GetNumRows();
      UInt_t row = 0;
      while (row < numRows)
         {
         TFElementPtr & elPtr  = col[row];
         Bool_t isRelative = elPtr.IsRelativePath();

         // absolute path of new input file
         TFFilePath inAbsName(elPtr.GetFileName());
         inAbsName.MakeAbsolutePath(inGroupName.Data());
         if (params.logLevel & 2)
            printf("converting  %s[%s]\n", inAbsName.Data(), elPtr.GetElementName());  
         TFIOElement * element = TFRead(inAbsName.Data(), elPtr.GetElementName(), 
                                        elPtr.GetCycle());
         if (element == NULL)
            {
            // delete this row
            group->DeleteRows(1, row);
            numRows--;
            continue;
            }

         // relative path to its group
         TFFilePath relName(elPtr.GetFileName());
         if (!isRelative)
            relName.MakeRelativePath(inGroupName.Data());

         // absolute path of output file
         TFFilePath outAbsName(relName.Data());
         outAbsName.MakeAbsolutePath(outGroupName.Data());

         if (element->IsA() == TFGroup::Class())
            ConvertGroup((TFGroup*)element, inAbsName, outAbsName, params, 
                         count);
                                                
        
         // save the element in the new file
         Int_t cycle;
         Bool_t saved = kFALSE;
         try
            {
            const char * newFileName = ConvertExtension(outAbsName.Data(), params);
            if (!ElementExist(newFileName, element, &cycle))
               {
               MakeDir(ConvertExtension(outAbsName.Data(), params));
               element->SaveElement(ConvertExtension(outAbsName.Data(), params), 
                                    params.compLevel);
               cycle = element->GetCycle();
               saved = kTRUE;
               }
            }
         catch (TFException) {}

         if (saved && !element->IsFileConnected())
            {
            // delete this row
            group->DeleteRows(1, row);
            delete element;
            numRows--;
            continue;
            }
         
         // update the element pointer in this column
         try
            {
            TFDataType dataType;
            if (element->IsA()->InheritsFrom(TFGroup::Class()))
               dataType = kGroupType;
            else if (element->IsA()->InheritsFrom(TFTable::Class()))
               dataType = kTableType;
            else if (element->IsA()->InheritsFrom(TFBaseImage::Class()))
               dataType = kImageType;
            else
               dataType = kBaseType;

            if (isRelative)
               {
               col[row] = TFElementPtr(ConvertExtension(elPtr.GetFileName(), params), 
                                       element->GetName(), cycle, dataType);
               }
            else
               col[row] = TFElementPtr(ConvertExtension(outAbsName.Data(), params), 
                                       element->GetName(), cycle, dataType);

            count.conv++;
            }
         catch (TFException) {}

         delete element;
         row++;
         }

      count.groups++;
      }
      catch (TFException) {}
}
//_____________________________________________________________________________
static const char * GetOutFile(Params & params, const char * inFile)
{
//  depending on the program parameters (sameDir and outDir) this functions
//  creates the new full path of the top level file. It does not convert the
//  extension.

   if (params.sameDir)
      return inFile;

   static char out[L_PATH];

   strcpy(out, params.outDir);
   const char * pos = strrchr(inFile, DS);
   if (pos)
      strcat(out, pos);
   else
      {
      strcat(out, DS2);
      strcat(out, inFile);
      }
   return out;
}
//_____________________________________________________________________________
static char * strrstr(char * str, const char * sub)
{
//  very similar as strstr, but finds the last occurrence of sub in str

   char * pos = strstr(str, sub);
   char * prev = NULL;

   while (pos)
      {
      prev = pos;
      pos = strstr(pos + 1, sub);
      }
   return prev;
}

//_____________________________________________________________________________
static const char * ConvertExtension(const char * fileName, Params & params)
{
//  Converts the extension of the input file name into the extension
//  of all output files (defined by a program parameter)
//  Returns the new filename. 
//  The pointer to the converted file name (return value) always points
//  to the same string. Therefore it will be overwritten at the next call
//  of this function!
// 
//  function throws a TFException exception if the input file does not 
//  have one of the expected extensions: fit*, root or gadc

   static char out[L_PATH];

   strcpy(out, fileName);
   char * pos = strrstr(out, ".fit");
   if (pos == NULL)
      {
      pos = strrstr(out, ".root");
      if (pos == NULL)
         pos = strrstr(out, ".asro");
         if (pos == NULL)
            throw TFException("ConvertExtension", 
                              "Unknown file extension: %s", fileName);
      }
   *(pos + 1) = 0;

   strcat(out, params.fileFormat);

   return out;
}
//_____________________________________________________________________________
static bool ElementExist(const char * fileName, TFIOElement * element, 
                         Int_t * cycle)
{
//   returns true if an identical element as "element" exist already in
//   the file fileName

   TFError::SetErrorType(kNoErr);
     
   TFFileIter fl(fileName);
   TFError::SetErrorType(kAllErr);

   while (fl.Next())
      {
      if (element->IsA() == fl->IsA() &&
          *element == *fl)
         {
         *cycle = fl->GetCycle();
         return true;
         }
      }

   return false;
}
/*****************************************************************************/
/*  Function :     MakeDir                                                   */
/*                                                                           */
/*  Description:   creates if necessary the directories to write the file    */
/*                 inName.                                                   */
/*                                                                           */
/*  Return code:   none, may throw an TFException exception                  */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  inName      input   filename         filename with full path. The path   */
/*                                       will be created if necessary        */
/*                                                                           */
/*****************************************************************************/
static void MakeDir(const char * inName)
{

	char * name, *end;

	/* copy the name to a new data buffer. We will change it during processing*/
	name = (char*)malloc(strlen(inName) + 1);
	strcpy(name, inName);

   // remove // 
   while (char * pos = strstr(name, "//"))
      memmove(pos, pos+1, strlen(pos));

	/* find the end of the directory path                                     */
	end = strrchr(name, DS);
	if (end != NULL )
		{
		/* there are parent directories defined in our path. Check if they     */
		/* already exist if not create them.                                   */
		*end = 0;
		MakeDirRecursive(name);
		}

	free(name);

}

/*****************************************************************************/
/*  Function :     MakeDirRecursive                                          */
/*                                                                           */
/*  Description:   creates the top directory of name. Calls itself if parent */
/*                 directories has to be created.                            */
/*                                                                           */
/*  Return code:   none, may throw an TFException exception                  */
/*                                                                           */
/*  Parameter arguments:                                                     */
/*                                                                           */
/*  name        I/O     valid range      description                         */
/*                                                                           */
/*  name        input   dir path         directroy path that has to be       */
/*                                       created if not already exist.       */
/*                                                                           */
/*****************************************************************************/
static void MakeDirRecursive(char * name)
{

	struct stat dirStatus;
	int         rc;
	char        * end;

	/* check if name already exist                                            */
	rc = stat(name, &dirStatus);
	if (rc == 0)
		{
		/* the file already exist. Make sure it is an directory                */
		if (!(dirStatus.st_mode & S_IFDIR))
         throw TFException("MakeDirRecursive", 
                          "Cannot create directroy %s. "
                          "File with same name already exist", name);

		/* there is nothing to do any more                                     */
		return;
		}

	/* the directory does not exist                                           */
	end = strrchr(name, DS);
	if (end)
		{
		/* there are parent directories defined in our path. Check if they     */
		/* already exist if not create them.                                   */
		*end = 0;
		MakeDirRecursive(name);
		*end = DS;
		}
	
	/* now we are ready to make a new directory                            */
	rc = mkdir(name, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
	if (rc != 0)
      throw TFException("MakeDirRecursive", 
                        "Error during creating directroy %s", name);

	return;
}

