// ////////////////////////////////////////////////////////////////////////////
//
//  File:      tfdump.cxx
//
//  Version:   1.1
//
//  Author:    Reiner Rohlfs (GADC)
//
//  History:   1.0   16.07.03  first released version
//             1.1   11.08.04  mumber of bins per row are printerd for each
//                             column
//
// ////////////////////////////////////////////////////////////////////////////
#include <stdio.h>

#include "TFIOElement.h"
#include "TFError.h"
#include "TFGroup.h"
#include "TFColumn.h"
#include "TFImage.h"

#include "pil.h"
#include "param_api.h"

#include "tfdump.h"

//_____________________________________________________________________________
static void Dump1Column(TFBaseCol & col, Params * params)
{

   // dump basic information of column
   UInt_t numRows = col.GetNumRows();
   Int_t  numBins = col.GetNumBins();
   printf("name of column: %-20s  unit: %-6s  data type: %s (%d byte)\n", 
            col.GetName(), col.GetUnit(), col.GetTypeName(),
            col.GetWidth());
   printf("%36s  number of bins per row: %2d\n", "",  col.GetNumBins());

   // the header of the column if wanted
   if (params->header)
      {
      TFAttrIter i_header = col.MakeAttrIterator();
      char hstr[500];
      while (i_header.Next())
         {
         printf("    %-16s : %20s %-5s | %s\n",
                  i_header->GetName(),
                  i_header->GetStringValue(hstr),
                  i_header->GetUnit(),
                  i_header->GetComment() );
         }
      }

   // the data of the column if wanted
   if (params->data)
      {
      char hstr[500];
      for (UInt_t row = 0; row < numRows; row++)
         {
         printf("%3d ", row);
         for (Int_t bin = 0; bin < numBins; bin++)
            if (col.IsNull(row, bin))
               printf("NULL    ");
            else
               printf("%s ", col.GetStringValue(row, bin, hstr, -8));

         printf("\n");
         }
      printf("\n");
      }
}
//_____________________________________________________________________________
static void DumpColumns(TFTable * table, Params * params)
{
   if (params->columns[0] == 0)
      {
      // dump all columns
      TFColIter i_col = table->MakeColIterator();
      while(i_col.Next())
         Dump1Column(*i_col, params);
      }
   else
      {
      // dump requested columns
      try {
         TFBaseCol & col = table->GetColumn(params->columns);
         Dump1Column(col, params);
         }
      catch (TFException) { }
      }

}
//_____________________________________________________________________________
static void DumpImage(TFBaseImage * element)
{
   UInt_t numDim = element->GetNumDim();
   printf("Number of dimensions: %u\n", numDim);
   UInt_t * size = new UInt_t[numDim];
   element->GetSize(size);

   if (numDim == 1)
      printf("X: %u", size[0]);
   else if (numDim == 2)
      printf("Y: %u  X: %u", size[0], size[1]);
   else if (numDim == 3)
      printf("Z: %u  Y: %u  X: %u", size[0], size[1], size[2]);
   else
      for (int dim = 0; dim < numDim; dim++)
         printf("dim %d: %u  ", dim + 1, size[dim]);
   printf("\n");

   delete [] size;
}
//_____________________________________________________________________________
static void Dump1Element(TFIOElement * element, Params * params)
{
   // a new element
   printf("=============================================="
          "==================================\n");
 
   // basic information of the element
   printf("type of element: ");
   if (element->IsA() == TFIOElement::Class())
      printf("%-16s", "io - element");
   else if (element->IsA() == TFTable::Class())
      printf("%-16s  number of rows: %u  number of columns: %u", "table", 
      ((TFTable*)element)->GetNumRows(), ((TFTable*)element)->GetNumColumns() );
   else if (element->IsA() == TFGroup::Class())
      printf("%-16s  number of rows: %u  number of columns: %u", "group", 
      ((TFGroup*)element)->GetNumRows(), ((TFGroup*)element)->GetNumColumns() );
   else if (element->IsA()->InheritsFrom(TFBaseImage::Class()) )
      {
      printf("%-16s", element->IsA()->GetName());
      }
   else 
      printf("%-16s", "unknown");

   printf("\nname of element: %s\ncycle number   : %d\n\n", 
           element->GetName(), element->GetCycle());

   // the header of the element if wanted
   if (params->header)
      {
      TFAttrIter i_header = element->MakeAttrIterator();
      char hstr[500];
      while (i_header.Next())
         {
         printf("%-16s : %20s %-5s | %s\n",
                i_header->GetName(),
                i_header->GetStringValue(hstr),
                i_header->GetUnit(),
                i_header->GetComment() );
         }
      }

   // dump further data depending on the type of element
   if (element->IsA() == TFTable::Class() ||
       element->IsA() == TFGroup::Class()   )
      DumpColumns((TFTable*)element, params);
   else if (element->IsA()->InheritsFrom(TFBaseImage::Class()) )
      DumpImage((TFBaseImage*)element);

   printf("\n");
}

//_____________________________________________________________________________
static int ReadParams(Params * params, int argc, char ** argv)
{
   int status = ISDC_OK;
 
   for (int par = 1; par < argc; par++)
      if (strcmp("--g",   argv[par]) == 0 ||
          strcmp("--gui", argv[par]) == 0    )
         {
         // use the GUI to ask for the parameters
         ParamGui gui(argc, argv);
         if (gui.Quit())        exit(0);
         if (gui.Error() != ISDC_OK)   return gui.Error();
      
         strcpy(params->file,       gui.GetString("file"));
         strcpy(params->name,       gui.GetString("name"));
         strcpy(params->columns,    gui.GetString("column"));
         params->header =   gui.GetBool("header");
         params->cycle =   gui.GetInt("cycle");
         params->data =    gui.GetBool("data");

         return gui.Error();
         }

   do
      {
      status = PILInit(argc, argv);
      if (status != ISDC_OK) break;

      status = PILGetFname("file", params->file);
      if (status != ISDC_OK) break;

      status = PILGetBool("header", &params->header);
      if (status != ISDC_OK) break;
      
      status = PILGetString("name", params->name);
      if (status != ISDC_OK) break;
      
      if (params->name[0] != 0)
         {
         status = PILGetInt("cycle", &params->cycle);
         if (status != ISDC_OK) break;
         }

      status = PILGetString("column", params->columns);
      if (status != ISDC_OK) break;

      status = PILGetBool("data", &params->data);
      if (status != ISDC_OK) break;

      } while (0);

   PILClose(ISDC_OK);

   return status;
}

//_____________________________________________________________________________
main(int argc, char * argv[])
{
   Params params;

   int status = ReadParams(&params, argc, argv);

   if (params.file[0] == 0)
      {
      printf("GUI not yet implemented\n");
      return 0;
      }

   printf("\ndump of file %s\n\n", params.file);

   TFError::SetErrorType(kAllErr);

   try {
      if (params.name[0] == 0)
         {
         // dump all elements in file
         TFFileIter fl(params.file);

         while (fl.Next())
            Dump1Element(&*fl, &params);

         }   
      else
         {
         // dump one element
         TFIOElement * element = TFRead(params.file, params.name, params.cycle);
         if (element)
            Dump1Element(element, &params);

         delete element;
         }

      }
   catch (TFException) { }

   TFError::PrintErrors();

}
